from xslim.logger import logger


# attribute checker and preprocess
def process_attribute(attr, input_shape, kernel_shape=None, op_type=None):
    # ASSUME input is 2D
    # assert len(input_shape) == 2
    # Get default attr value
    auto_pad = attr.get("auto_pad", "NOTSET")
    strides = attr.get("strides", [1, 1])
    dilations = attr.get("dilations", [1, 1])
    kernels = attr.get("kernel_shape", kernel_shape)
    pad_needed = None

    if op_type == "ConvTranspose" and "output_shape" in attr:
        output_shape = attr["output_shape"]
        out_pad = [0, 1] if output_shape % 2 != 0 else [0, 0]
        pad_needed = [
            (input_shape[i] - 1) * strides[i] + dilations[i] * (kernels[i] - 1) + 1 + out_pad[i] - output_shape[i]
            for i in range(len(input_shape))
        ]

    if auto_pad != "NOTSET":
        if "pads" in attr:
            logger.warning("auto_pad is conflict with pads attribute. Use pads here.")
        elif auto_pad == "VALID":
            attr["pads"] = [0, 0, 0, 0]
        elif auto_pad in ("SAME_UPPER", "SAME_LOWER"):
            if op_type == "ConvTranspose":
                # `output_padding` is only used to find output shape, but does not actually add zero-padding to output
                out_pad = attr.get("output_padding", [0, 0])
                output_shape = [input_shape[i] * strides[i] for i in range(len(input_shape))]
                pad_needed = [
                    (input_shape[i] - 1) * strides[i]
                    + dilations[i] * (kernels[i] - 1)
                    + 1
                    + out_pad[i]
                    - output_shape[i]
                    for i in range(len(input_shape))
                ]
            else:
                output_shape = [(input_shape[i] + strides[i] - 1) // strides[i] for i in range(len(input_shape))]
                pad_needed = [
                    (output_shape[i] - 1) * strides[i] + dilations[i] * (kernels[i] - 1) + 1 - input_shape[i]
                    for i in range(len(input_shape))
                ]
        else:
            raise ValueError(f"Invalid auto_pad value {auto_pad}")

    if pad_needed is not None:
        pads = []
        for item in pad_needed:
            pads.append((item if auto_pad == "SAME_UPPER" else item + 1) // 2)
        # onnx pads format should be as follow [x1_begin, x2_begin...x1_end, x2_end,...]
        pads = pads + [pad_needed[i] - p for i, p in enumerate(pads)]
        attr["pads"] = pads
        # onnx pads attribute cannot be used simultaneously with auto_pad attribute
        attr.pop("auto_pad")


def preprocess_attr(attr, op_type=None):
    processed_attribute = {}
    if "kernel_shape" in attr and op_type == "Pooling":
        processed_attribute["kernel_size"] = attr["kernel_shape"]
    if "group" in attr:
        processed_attribute["groups"] = attr["group"]
    if "pads" in attr:
        # Change pads from start-end to torch format
        pads = attr["pads"]
        assert len(pads) % 2 == 0
        if len(pads) == 4:
            begin_pad = pads[:2]
            end_pad = pads[2:]
            if begin_pad == end_pad:
                processed_attribute["padding"] = begin_pad
            else:
                raise ValueError("Torch function only support begin_pad == end_pad in layer")
        else:
            processed_attribute["padding"] = pads

    if "dilations" in attr:
        processed_attribute["dilation"] = attr["dilations"]
    if "strides" in attr:
        processed_attribute["stride"] = attr["strides"]
    if "ceil_mode" in attr:
        processed_attribute["ceil_mode"] = bool(attr["ceil_mode"])
    return processed_attribute



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import math
from typing import Callable, Dict, Iterable, List, Sequence, Set, Tuple, Union

import torch
from xslim.logger import logger

from ..defs import (
    COMPUTING_OP,
    OBSERVER_MAX_BIAS_VAL,
    OBSERVER_MIN_SCALE_THRESHOLD,
    OBSERVER_SIGMOID_MAX_VALUE,
    PASSIVE_OPERATIONS,
    XQUANT_CONFIG,
)
from ..ppq_decorator import (
    BaseGraph,
    Operation,
    QuantableOperation,
    QuantizationOptimizationPass,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    SearchableGraph,
    Variable,
    convert_any_to_torch_tensor,
    empty_ppq_cache,
    minmax_to_scale_offset,
    ppq_common,
)


class PassiveParameterBakingPass(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name="XSlim PassiveParameterBakingPass Pass")

    @staticmethod
    def passive_parameters_quant(op: QuantableOperation):
        def check_state(state: QuantizationStates):
            return state in {
                QuantizationStates.PASSIVE,
                QuantizationStates.ACTIVATED,
                QuantizationStates.BAKED,
                QuantizationStates.OVERLAPPED,
            }

        if not isinstance(op, QuantableOperation):
            return

        if op.type in {"Clip"}:
            # inputs are [input value, min[optional], max[optional]]
            i_cfg = op.config.input_quantization_config[0]
            if not check_state(i_cfg.state):
                return
            for config in op.config.input_quantization_config[1:]:
                config.master_by = i_cfg
                config.visibility = QuantizationVisibility.INTERNAL
        elif op.type in {"Pad"}:
            # inputs are [input value, pad[shape-related], pad value[optional]]
            if op.num_of_input != 3:
                return
            i_cfg = op.config.input_quantization_config[0]
            if not check_state(i_cfg.state):
                return
            if len(op.config.input_quantization_config) > 1:
                pad_config = op.config.input_quantization_config[-1]
                pad_config.master_by = i_cfg
                pad_config.visibility = QuantizationVisibility.INTERNAL
        elif op.type not in COMPUTING_OP:
            for in_config, in_var in op.config_with_variable:
                if (
                    in_var.is_parameter
                    and in_config.state == QuantizationStates.INITIAL
                    and in_config.policy.has_property(QuantizationProperty.ASYMMETRICAL)
                    and in_config.policy.has_property(QuantizationProperty.PER_TENSOR)
                ):
                    max_range_val = float(in_var.value.max())
                    min_range_val = float(in_var.value.min())
                    if torch.all(in_var.value.to(torch.int8) - in_var.value == 0):
                        scale = torch.tensor(1.0, dtype=torch.float32, device=in_var.value.device)
                        offset = torch.tensor(
                            math.ceil((in_config.quant_max + in_config.quant_min) / 2),
                            dtype=torch.float32,
                            device=in_var.value.device,
                        )
                    elif torch.all(in_var.value.to(torch.uint8) - in_var.value == 0):
                        scale = torch.tensor(1.0, dtype=torch.float32, device=in_var.value.device)
                        offset = torch.tensor(
                            in_config.quant_min,
                            dtype=torch.float32,
                            device=in_var.value.device,
                        )
                    elif min_range_val != max_range_val:
                        scale, offset = minmax_to_scale_offset(
                            min_range_val, max_range_val, in_config, OBSERVER_MIN_SCALE_THRESHOLD
                        )
                    elif max_range_val > 0:
                        scale, offset = minmax_to_scale_offset(
                            0, max_range_val, in_config, OBSERVER_MIN_SCALE_THRESHOLD
                        )
                    elif max_range_val < 0:
                        scale, offset = minmax_to_scale_offset(
                            max_range_val, 0, in_config, OBSERVER_MIN_SCALE_THRESHOLD
                        )
                    else:
                        continue
                    in_config.scale = convert_any_to_torch_tensor(scale)
                    in_config.offset = convert_any_to_torch_tensor(offset)
                    in_config.state = QuantizationStates.PASSIVE

    @staticmethod
    def passive_bias_quant(operation: QuantableOperation):
        if not isinstance(operation, QuantableOperation):
            return
        if operation.type not in {"Conv", "ConvTranspose", "Gemm"}:
            return
        if operation.num_of_input == 3:
            i_cfg, w_cfg, b_cfg = operation.config.input_quantization_config
            o_cfg = operation.config.output_quantization_config[0]
            if (
                b_cfg.state not in {QuantizationStates.PASSIVE_INIT}
                or not operation.inputs[1].is_parameter
                or not operation.inputs[-1].is_parameter
                or w_cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL)
                or b_cfg.policy.has_property(QuantizationProperty.ASYMMETRICAL)
            ):
                return
            bias = operation.inputs[-1].value
            assert bias.numel() == bias.shape[-1], (
                f"For op {operation.name}, expect Bias shape to be {[bias.numel()]}, " f"however {bias.shape} was given"
            )
            operation.inputs[-1].value = bias.squeeze()
            if operation.inputs[-1].value.ndim == 0 and operation.inputs[-1].value.numel() == 1:
                operation.inputs[-1].value = operation.inputs[-1].value.unsqueeze(0)
            if w_cfg.scale is None or i_cfg.scale is None:
                return

            if w_cfg.scale.numel() > 1:
                if operation.type in {"Conv"}:
                    iw_zero_channel = torch.where(
                        w_cfg.scale * i_cfg.scale <= (OBSERVER_MIN_SCALE_THRESHOLD * 2**-7)
                    )[0]
                    w_zero_channel = torch.where(w_cfg.scale <= OBSERVER_MIN_SCALE_THRESHOLD)[0]
                    if iw_zero_channel.numel() > 0 or w_zero_channel.numel() > 0:
                        operation.inputs[1].value[iw_zero_channel] = 0
                        operation.inputs[1].value[w_zero_channel] = 0
                        store_state = w_cfg.state
                        w_cfg.state = QuantizationStates.INITIAL
                        w_cfg.scale[iw_zero_channel] = 1.0
                        w_cfg.scale[w_zero_channel] = 1.0
                        w_cfg.state = store_state

            _b_scale = w_cfg.scale * i_cfg.scale
            _i_bias = bias.to(torch.float64) / _b_scale.to(torch.float64)
            b_cfg.scale = _b_scale
            if operation.type in {"Conv"} and operation.attributes.get("group", 1) == 1:
                if torch.any(torch.abs(_i_bias) > OBSERVER_MAX_BIAS_VAL):
                    b_cfg.scale = o_cfg.scale
                    b_cfg.state = QuantizationStates.FP32
                    return
            b_cfg.state = QuantizationStates.PASSIVE
            b_cfg.offset = torch.zeros_like(b_cfg.scale)

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for _, operation in graph.operations.items():
            PassiveParameterBakingPass.passive_parameters_quant(operation)
            PassiveParameterBakingPass.passive_bias_quant(operation)


class AsymmetricaUnsignlAlignSign(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name="XSlim AsymmetricalAlignS8 Pass")

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for _, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue
            for config, var in [_ for _ in operation.config_with_variable]:
                if config.policy.has_property(QuantizationProperty.ASYMMETRICAL) and config.quant_min == 0:
                    config.quant_min = -(2 ** (config.num_of_bits - 1))
                    config.quant_max = 2 ** (config.num_of_bits - 1) - 1
                    if config.dominated_by == config and config.offset is not None:
                        store_state = config.state
                        config.state = QuantizationStates.INITIAL
                        config.offset = config.offset + config.quant_min
                        config.state = store_state


class QuantizeFusionPass(QuantizationOptimizationPass):
    def __init__(
        self,
        fuse_relu_clip: bool = True,
        fuse_passive_op: bool = True,
    ) -> None:
        self.fuse_relu_clip = fuse_relu_clip
        self.fuse_passive_op = fuse_passive_op
        super().__init__(name="XSlim Quantization Fusion Pass")

    def is_same_platform(self, operations: List[Operation]):
        platforms = [operation.platform for operation in operations]
        return all([platform == platforms[0] for platform in platforms])

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        processor = SearchableGraph(graph)

        if self.fuse_passive_op:
            # all passive operations should never changes quantization configuration of its input
            # so to say their input and output share a same scale.
            for op in graph.operations.values():
                if op.type not in PASSIVE_OPERATIONS:
                    continue
                source_op = op.inputs[0].source_op
                if source_op is None:
                    continue  # beginning op, can not merge.
                if isinstance(op, QuantableOperation) and self.is_same_platform([op, source_op]):
                    TQC = op.config.input_quantization_config[0]
                    for output_cfg in op.config.output_quantization_config:
                        output_cfg.dominated_by = TQC

        if self.fuse_relu_clip:
            patterns = processor.pattern_matching(
                patterns=[lambda x: True, lambda x: x.type in {"Relu", "Clip"}], edges=[[0, 1]], exclusive=True
            )
            for computing_op, act_op in patterns:
                if not isinstance(act_op, QuantableOperation) or not isinstance(computing_op, QuantableOperation):
                    continue

                if (
                    len(graph.get_downstream_operations(computing_op)) == 1
                    and len(graph.get_upstream_operations(act_op)) == 1
                ):
                    computing_op.config.output_quantization_config[0].dominated_by = (
                        act_op.config.output_quantization_config[0]
                    )
                    act_op.config.input_quantization_config[0].dominated_by = act_op.config.output_quantization_config[
                        0
                    ]


class ActivationClipRefine(QuantizationOptimizationPass):
    def __init__(self) -> None:
        super().__init__(name="XSlim ActivationClipRefine Pass")
        self.act_op_set = {"HardSigmoid", "Sigmoid"}

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for _, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue

            for tqc, var in operation.config_with_variable:
                if tqc.state in {QuantizationStates.INITIAL}:
                    if any([dest_op.type in self.act_op_set for dest_op in var.dest_ops]):
                        force_range_min = None
                        force_range_max = None
                        check_hardsigmoid = all([dest_op.type in {"HardSigmoid"} for dest_op in var.dest_ops])
                        check_hardswish = all([dest_op.type in {"HardSigmoid", "Mul"} for dest_op in var.dest_ops])

                        check_sigmoid = all([dest_op.type in {"Sigmoid"} for dest_op in var.dest_ops])
                        check_swish = all([dest_op.type in {"Sigmoid", "Mul"} for dest_op in var.dest_ops])

                        if check_hardsigmoid:
                            alpha = var.dest_ops[0].attributes.get("alpha", 0.1666666716337204)
                            beta = var.dest_ops[0].attributes.get("beta", 0.5)
                            force_range_min = -beta / alpha
                            force_range_max = (1.0 - beta) / alpha

                        elif check_sigmoid or (check_swish and len(var.dest_ops) == 2):
                            force_range_min = -OBSERVER_SIGMOID_MAX_VALUE
                            if check_sigmoid:
                                force_range_max = OBSERVER_SIGMOID_MAX_VALUE

                        elif check_hardswish and len(var.dest_ops) == 2:
                            dest_op_type = [dest_op.type for dest_op in var.dest_ops]
                            if "HardSigmoid" in dest_op_type and "Mul" in dest_op_type:
                                act_index = dest_op_type.index("HardSigmoid")
                                mul_index = dest_op_type.index("Mul")
                                if var.dest_ops[act_index].outputs[0] in var.dest_ops[mul_index].inputs:
                                    alpha = var.dest_ops[act_index].attributes.get("alpha", 0.1666666716337204)
                                    beta = var.dest_ops[act_index].attributes.get("beta", 0.5)
                                    force_range_min = -beta / alpha

                        if force_range_min is not None or force_range_max is not None:
                            operation._detail["output_force_range"] = {
                                var.name: {"min": force_range_min, "max": force_range_max}
                            }
                            operation._detail["output_force_range"].update({var.name: {"min": force_range_min}})


class QuantizeConfigRefinePass(QuantizationOptimizationPass):
    def __init__(
        self, precision_level: int = 0, custom_setting: Sequence["CustomQuantizationParameterSetting"] = None
    ) -> None:
        super().__init__(name="XSlim QuantizeConfigRefine Pass")
        self._precision_level = precision_level
        self._max_bits = XQUANT_CONFIG.max_bits
        self._quant_max = 2 ** (self._max_bits - 1) - 1
        self._quant_min = -(2 ** (self._max_bits - 1))
        self._custom_setting = custom_setting

    def precesion_level_2(self, graph: BaseGraph, operation: QuantableOperation):
        upstream_operations = graph.get_upstream_operations(operation)
        downstream_operations = graph.get_downstream_operations(operation)
        if (operation.type in {"Mul", "Add", "Sub", "Div"} and operation.num_of_parameter == 0) or operation.type in {
            "Softmax",
            "Sigmoid",
        }:
            self.precesion_level_3(graph, operation)

    def precesion_level_3(self, graph: BaseGraph, operation: QuantableOperation):
        for tqc, var in operation.config_with_variable:
            if tqc.dominated_by is tqc and tqc.state == QuantizationStates.INITIAL:
                tqc.state = QuantizationStates.FP32

    def custom_tqc_set(self, graph: BaseGraph, custom_tqc: "CustomQuantizationParameterSetting"):
        var_dict = graph.variables
        input_names = custom_tqc.input_names
        output_names = custom_tqc.output_names
        precision_level = custom_tqc.precision_level
        max_percentile = custom_tqc.max_percentile
        calibration_type = custom_tqc.calibration_type

        input_tensors = []
        for var_name in input_names:
            var = graph.variables.get(var_name, None)
            if var is None:
                raise ValueError("var name {} not in the graph.".format(var_name))
            if not var.is_parameter:
                input_tensors.append(var)

        output_tensors = []
        for var_name in output_names:
            var = graph.variables.get(var_name, None)
            if var is None:
                raise ValueError("var name {} not in the graph.".format(var_name))
            output_tensors.append(var)

        if len(input_tensors) < 1 or len(output_tensors) < 1:
            raise ValueError("input_names and output_names should be set.")

        visited_ops = []
        output_tensors_set = set(output_tensors)

        def get_op_blocks(in_vars: Sequence[Variable]):
            for var in in_vars:
                if var in output_tensors_set:
                    return
                for op in var.dest_ops:
                    visited_ops.append(op)
                    get_op_blocks(op.outputs)

        get_op_blocks(input_tensors)

        for op in visited_ops:
            if not isinstance(op, QuantableOperation):
                continue
            logger.info("set custom tqc for operation: {}".format(op.name))
            if isinstance(precision_level, int) and precision_level == 2:
                self.precesion_level_2(graph, op)
            elif isinstance(precision_level, int) and precision_level == 3:
                self.precesion_level_3(graph, op)
            tqc = op.config

            for in_var, in_tqc in zip(op.inputs, tqc.input_quantization_config):
                if not in_var.is_parameter:
                    if isinstance(max_percentile, float):
                        in_tqc.detail[ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE] = max_percentile
                    if isinstance(calibration_type, str):
                        in_tqc.observer_algorithm = calibration_type
            for out_var, out_tqc in zip(op.outputs, tqc.output_quantization_config):
                if not out_var.is_parameter:
                    if isinstance(max_percentile, float):
                        out_tqc.detail[ppq_common.OBSERVER_PERCENTILE_MANUL_OVERRIDE] = max_percentile
                    if isinstance(calibration_type, str):
                        out_tqc.observer_algorithm = calibration_type

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        sorted_ops = graph.topological_sort()
        if self._precision_level == 2:
            for operation in sorted_ops:
                if not isinstance(operation, QuantableOperation):
                    continue
                self.precesion_level_2(graph, operation)
        elif self._precision_level == 3:
            for operation in sorted_ops:
                if not isinstance(operation, QuantableOperation):
                    continue
                self.precesion_level_3(graph, operation)

        if isinstance(self._custom_setting, Sequence):
            for tqc_setting in self._custom_setting:
                self.custom_tqc_set(graph, tqc_setting)



"""PPQ Core File System PPQ 核心文件系统(IO)

You are not allowed to modify this 请勿修改此文件
"""
import os
from typing import Any

import numpy as np
import torch

from .config import PPQ_CONFIG
from .data import DataType, convert_any_to_numpy, convert_any_to_torch_tensor
from .defs import ppq_file_io, ppq_warning
import pickle


def is_file_exist(file: str):
    if os.path.exists(file):
        return os.path.isfile(file)
    else: return False


@ ppq_file_io
def open_txt_file_from_writing(file: str, mode: str, encoding: str='utf-8'):
    if is_file_exist(file): raise IOError('Writing File ')


class Serializable():
    """An interface which means a class instance is binary serializable,
    nothing funny."""
    def __init__(self) -> None:
        self._export_value = PPQ_CONFIG.DUMP_VALUE_WHEN_EXPORT

    def __setstate__(self, state: dict):
        if not isinstance(state, dict):
            raise TypeError(f'PPQ Data Load Failure. Can not load data from {type(state)}, '
                'Your data might get damaged.')

        if '__version__' not in state or state['__version__'] != PPQ_CONFIG.VERSION:
            ppq_warning(
                'You are loading an object created by PPQ with different version,'
                ' it might cause some problems.')

        for key, value in state.items():
            self.__dict__[key] = value
            if isinstance(value, ValueState):
                self.__dict__[key] = value.unpack()
        return self

    def __getstate__(self) -> dict:
        attribute_dicts = self.__dict__
        attribute_dicts['__version__'] = PPQ_CONFIG.VERSION
        serialized = dict()

        for name, value in attribute_dicts.items():
            if isinstance(value, np.ndarray) or isinstance(value, torch.Tensor):
                if self._export_value is False: value = None
                serialized[name] = ValueState(value)
            else: serialized[name] = value
        return serialized


class ValueState(Serializable):
    def __init__(self, value: Any) -> None:
        self._value_type = str(value.__class__.__name__)
        if isinstance(value, np.ndarray):
            self._dtype = value.dtype
            self._shape = value.shape
            self._value = pickle.dumps(value)
        elif isinstance(value, torch.Tensor):
            self._dtype = DataType.to_numpy(DataType.convert_from_torch(value.dtype))
            self._shape = value.shape
            self._value = pickle.dumps(convert_any_to_numpy(value))
        elif isinstance(value, list) or isinstance(value, tuple):
            self._value = value
            self._dtype = None
            self._shape = None
        elif value is None:
            self._dtype = None
            self._shape = None
            self._value = None
        else:
            raise TypeError(f'PPQ Data Dump Failure, can not dump value type {type(value)}')

    def unpack(self) -> Any:
        if self._value_type == str(None.__class__.__name__):
            return None
        elif self._value_type == str('ndarray'):
            value = pickle.loads(self._value)
            assert isinstance(value, np.ndarray)
            value = value.astype(self._dtype)
            value = value.reshape(self._shape)
            return value
        elif self._value_type == str('Tensor'):
            if self._value is not None:
                value = pickle.loads(self._value)
                assert isinstance(value, np.ndarray)
                value = value.astype(self._dtype)
                if value is not None:
                    value = value.reshape(self._shape)
                value = convert_any_to_torch_tensor(
                    value, device='cpu', 
                    dtype=DataType.to_torch(DataType.convert_from_numpy(self._dtype)))
                return value
            else:
                return torch.tensor([], device='cpu')
        elif self._value_type in {'list', 'tuple', 'dict'}:
            return self._value



from abc import ABCMeta, abstractmethod
from ...core import TensorQuantizationConfig
from typing import Any, Callable


class BaseQuantFunction(Callable, metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def __call__(self, input_tensor: Any, quantization_config: TensorQuantizationConfig, **kwargs) -> Any:
        raise NotImplemented("Implement this first.")



"""End-to-end regression tests for the public quantization pipeline."""

import os
import shutil
import sys
import tempfile
import unittest
import warnings

import onnx
import torch
import torchvision
from onnx import TensorProto
from onnx.external_data_helper import convert_model_to_external_data

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

import xslim


class TestQuantizePipeline(unittest.TestCase):
    """Validate model-level FP16 and dynamic quantization flows."""

    input_shape = (1, 3, 64, 64)
    model_cache_dir = os.environ.get("XSLIM_TEST_MODEL_CACHE")
    _tempdirs = []

    @classmethod
    def _export_torchvision_model(cls, model, path):
        model.eval()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            torch.onnx.export(
                model,
                torch.randn(*cls.input_shape),
                path,
                input_names=["input"],
                output_names=["output"],
                opset_version=17,
                dynamo=False,
            )

    @classmethod
    def _get_model_path(cls, model_name):
        if cls.model_cache_dir:
            os.makedirs(cls.model_cache_dir, exist_ok=True)
            return os.path.join(cls.model_cache_dir, "{}.onnx".format(model_name))
        return None

    @classmethod
    def _prepare_model(cls, model_name, model):
        cached_model_path = cls._get_model_path(model_name)
        if cached_model_path is not None:
            if not os.path.exists(cached_model_path):
                cls._export_torchvision_model(model, cached_model_path)
            return cached_model_path

        tempdir = tempfile.mkdtemp()
        cls._tempdirs.append(tempdir)
        model_path = os.path.join(tempdir, "{}.onnx".format(model_name))
        cls._export_torchvision_model(model, model_path)
        return model_path

    @classmethod
    def tearDownClass(cls):
        """Remove any temporary directories created during tests."""
        for tempdir in getattr(cls, "_tempdirs", []):
            shutil.rmtree(tempdir, ignore_errors=True)

    @classmethod
    def _build_config(cls, model_path, precision_level):
        return {
            "model_parameters": {
                "onnx_model": model_path,
                "skip_onnxsim": True,
            },
            "calibration_parameters": {
                "input_parametres": [
                    {
                        "input_name": "input",
                        "input_shape": list(cls.input_shape),
                        "dtype": "float32",
                    }
                ]
            },
            "quantization_parameters": {
                "precision_level": precision_level,
                "analysis_enable": False,
            },
        }

    def test_dynamic_quantize_pipeline_supports_external_data_resnet18(self):
        with tempfile.TemporaryDirectory() as tempdir:
            source_model_path = self._prepare_model("resnet18", torchvision.models.resnet18(weights=None))
            model_path = os.path.join(tempdir, "resnet18_external.onnx")
            output_path = os.path.join(tempdir, "resnet18_external.dynq.onnx")

            shutil.copyfile(source_model_path, model_path)

            model_with_external_data = onnx.load(model_path)
            convert_model_to_external_data(
                model_with_external_data,
                all_tensors_to_one_file=True,
                location="resnet18_external.data",
                size_threshold=0,
                convert_attribute=False,
            )
            onnx.save_model(model_with_external_data, model_path)

            external_model = onnx.load(model_path, load_external_data=False)
            self.assertTrue(
                any(initializer.data_location == TensorProto.EXTERNAL for initializer in external_model.graph.initializer)
            )
            self.assertTrue(os.path.exists(os.path.join(tempdir, "resnet18_external.data")))

            quantized_model = xslim.quantize_onnx_model(
                self._build_config(model_path, precision_level=3), output_path=output_path
            )

            self.assertTrue(os.path.exists(output_path))
            onnx.checker.check_model(quantized_model)

            node_types = {node.op_type for node in quantized_model.graph.node}
            self.assertTrue(any("DynamicQuantize" in op_type for op_type in node_types))
            self.assertIn("DequantizeLinear", node_types)

    def test_fp16_pipeline_converts_mobilenet_v2_end_to_end(self):
        with tempfile.TemporaryDirectory() as tempdir:
            source_model_path = self._prepare_model("mobilenet_v2", torchvision.models.mobilenet_v2(weights=None))
            model_path = os.path.join(tempdir, "mobilenetv2.onnx")
            output_path = os.path.join(tempdir, "mobilenetv2.fp16.onnx")

            shutil.copyfile(source_model_path, model_path)

            fp16_model = xslim.quantize_onnx_model(
                self._build_config(model_path, precision_level=4), output_path=output_path
            )

            self.assertTrue(os.path.exists(output_path))
            onnx.checker.check_model(fp16_model)
            self.assertEqual(fp16_model.graph.input[0].type.tensor_type.elem_type, TensorProto.FLOAT)
            self.assertIn(TensorProto.FLOAT16, {initializer.data_type for initializer in fp16_model.graph.initializer})


if __name__ == "__main__":
    unittest.main()



from typing import Any, List, Tuple

from ..core import (
    OperationQuantizationConfig,
    QuantizationStates,
    TargetPlatform,
    TensorQuantizationConfig,
    convert_any_to_torch_tensor,
)
from ..quantization.qfunction import BaseQuantFunction

from .base.command import (
    GraphCommand,
    GraphCommandType,
    QuantizeOperationCommand,
    ReplaceOperationCommand,
    ReplaceVariableCommand,
)
from .base.graph import Operation, Variable
from .processer import GraphCommandProcessor


class QuantableOperation(Operation):
    """Quantable Operation (量化算子) 用来表示一个已经被量化了的算子
    相比于普通算子，一个量化算子具有以下额外的功能

    1. 每一个量化算子都将具有一个 config(OperationQuantizationConfig) 属性
       PPQ 使用这个东西描述量化细节，在整个网络中，有且只有这一个量化表示
       executor, dispatcher, optimization pass, exporter都是围绕这一属性工作的

    2. 每一个量化算子都将有一个 dequantize 方法和 restore_quantize_state 方法，
       一旦一个量化算子被 dequantize() 方法解除量化，该算子的 OperationQuantizationConfig 将被修改状态
       从而使得该算子的输入输出量化被暂时停用
       被解除量化的算子可以随时通过 restore_quantize_state 方法恢复量化状态
       对一个算子多次重复执行 dequantize 是可以的

    3. 每一个量化算子都将有一个 baking parameter 方法
       当算子具有有效的量化参数时，baking_parameters() 方法将对该算子的参数执行静态量化
       一旦静态量化完成，算子参数将被量化后的值替换；同时 config 的状态将被设置为: baked

    4. 每一个量化算子都将有一个 store_parameter_value 方法
       该方法将算子目前的参数保存入缓存；PPQ 将在创建 QuantableOperation 时执行此函数
       从而保存算子的原始参数，以备后续取用。
       一个显而易见的例子是，一旦算子执行了 baking_parameters 方法，它的参数值将被修改，
       此时若要完全还原算子状态，需要从缓存中取出算子的原始参数，并替换当前的值
       当你调用 restore_quantize_state 时，该方法会从缓存中取回保存的参数值并执行替换
       你不应当手动调用该方法，该方法将影响到 PPQ 的核心逻辑正确性

    5. 一个量化算子是可拷贝的，该拷贝只会拷贝算子的基本信息以及绑定的 OperationQuantizationConfig

    Args:
        Operation (_type_): _description_
    """

    def __init__(self, convert_from: Operation, quantize_config: OperationQuantizationConfig, platform: TargetPlatform):
        # Simply copy all attributes from fp32 operation
        # inputs, outputs will be created by QuantableGraph
        super().__init__(
            op_type=convert_from.type,
            inputs=convert_from.inputs.copy(),
            outputs=convert_from.outputs.copy(),
            attributes=convert_from.attributes,
            name=convert_from.name,
            platform=platform,
            opset=convert_from.opset,
        )

        self._config = quantize_config
        self._dequantized = False

    @property
    def config(self) -> OperationQuantizationConfig:
        return self._config

    @config.setter
    def config(self, config):
        """we will update variable's during this function.

        Args:
            config ([type]): [description]

        Raises:
            TypeError: [description]
            ExecError: [description]

        Returns:
            [type]: [description]
        """
        if not isinstance(config, OperationQuantizationConfig):
            raise TypeError(
                f"object {str(config)}({type(config)}) is not a acceptable config for operation {self.name}"
            )
        self._config = config

    @property
    def input_quant_config(self) -> List[TensorQuantizationConfig]:
        return self.config.input_quantization_config

    @property
    def output_quant_config(self) -> List[TensorQuantizationConfig]:
        return self.config.output_quantization_config

    def baking_parameters(self, quant_func: BaseQuantFunction):
        for config, var in self.config_with_variable:
            if var.is_parameter and config.state in {QuantizationStates.ACTIVATED, QuantizationStates.PASSIVE}:
                assert isinstance(var, QuantableVariable)
                assert len(var.dest_ops) == 1, (
                    f", Parameter {var.name} has {len(var.dest_ops)} destinations, "
                    "Baking parameter that has more than 1 destinations will incur unexpected problems, "
                    "PPQ does not support parameters with more than 1 related operation, reform your graph first."
                )
                var.value = quant_func(var.value, config)

                if config.state == QuantizationStates.ACTIVATED:
                    config.state = QuantizationStates.BAKED
                if config.state == QuantizationStates.PASSIVE:
                    config.state = QuantizationStates.PASSIVE_BAKED
        return self

    def store_parameter_value(self):
        for var, _ in zip(
            self.inputs + self.outputs, self.config.input_quantization_config + self.config.output_quantization_config
        ):
            if var.is_parameter:
                assert isinstance(var, QuantableVariable), "Unexpected error."
                # convert var.value to torch.Tensor
                # notice here we set device = None, this conversion will not change var.value.device anyway.
                # so that we can use var.value.device as a deploy device for stored_value
                var.stored_value = convert_any_to_torch_tensor(var.value, device="cpu").clone()
        return self

    def dequantize(self, parameter_only: bool = False, expire_device: str = "cpu"):
        if self._dequantized:
            return self
        for var, quant_config in zip(
            self.inputs + self.outputs, self.config.input_quantization_config + self.config.output_quantization_config
        ):
            if parameter_only and not var.is_parameter:
                continue
            quant_config.detail["Stored State"] = quant_config.state
            assert isinstance(var, QuantableVariable), f"Unexpected error with variable {var.name}."
            if var.is_parameter:
                # convert var.value to torch.Tensor
                # notice here we set device = None, this conversion will not change var.value.device anyway.
                # so that we can use var.value.device as a deploy device for stored_value
                stored_value = convert_any_to_torch_tensor(var.value, device=expire_device)
                var.value = convert_any_to_torch_tensor(var.value, device=None)
                var.value = convert_any_to_torch_tensor(
                    var.stored_value, device=var.value.device if var.value is not None else None
                )
                var.stored_value = stored_value
            quant_config.state = QuantizationStates.FP32
        self._dequantized = True
        return self

    def restore_quantize_state(self, expire_device: str = "cpu"):
        if not self._dequantized:
            return self
        for var, quant_config in zip(
            self.inputs + self.outputs, self.config.input_quantization_config + self.config.output_quantization_config
        ):
            if "Stored State" in quant_config.detail:
                quant_config.state = quant_config.detail["Stored State"]
                quant_config.detail.pop("Stored State")
                if var.is_parameter:
                    assert isinstance(var, QuantableVariable), "Unexpected error."
                    # convert var.value to torch.Tensor
                    # notice here we set device = None, this conversion will not change var.value.device anyway.
                    # so that we can use var.value.device as a deploy device for stored_value
                    stored_value = convert_any_to_torch_tensor(var.value, device=expire_device)
                    var.value = convert_any_to_torch_tensor(var.value, device=None)
                    var.value = convert_any_to_torch_tensor(
                        var.stored_value, device=var.value.device if var.value is not None else None
                    )
                    var.stored_value = stored_value
        self._dequantized = False
        return self

    @property
    def config_with_variable(self) -> List[Tuple[TensorQuantizationConfig, Variable]]:
        """Just a helper function, This function will list all related config
        and variable with current operation.

        Returns:
            List[Tuple[TensorQuantizationConfig, Variable]]: [description]
        """
        ret = []
        for cfg, var in zip(self.config.input_quantization_config, self.inputs):
            ret.append((cfg, var))
        for cfg, var in zip(self.config.output_quantization_config, self.outputs):
            ret.append((cfg, var))
        return ret

    def copy(self):
        return QuantableOperation(
            convert_from=super().copy(), quantize_config=self.config.copy(), platform=self.platform
        )


class QuantableVariable(Variable):
    def __init__(self, convert_from: Variable) -> None:
        super().__init__(
            name=convert_from.name,
            dest_ops=convert_from.dest_ops.copy(),
            source_op=convert_from.source_op,
            value=convert_from.value,
            is_parameter=convert_from.is_parameter,
            shape=convert_from.shape,
            dtype=convert_from.dtype,
        )
        self._fp32_value = None
        if convert_from.value is not None:
            self._fp32_value = convert_any_to_torch_tensor(convert_from.value, device="cpu")

    @property
    def stored_value(self) -> Any:
        return self._fp32_value

    @stored_value.setter
    def stored_value(self, value: Any):
        self._fp32_value = value

    @property
    def dest_op_configs(self) -> List[TensorQuantizationConfig]:
        _dest_op_configs, _dest_idx = [], self.dest_idx
        for idx, op in enumerate(self.dest_ops):
            if isinstance(op, QuantableOperation):
                _dest_op_configs.append(op.config.input_quantization_config[_dest_idx[idx]])
            else:
                _dest_op_configs.append(None)
        return _dest_op_configs

    @property
    def dest_op_platforms(self) -> List[TargetPlatform]:
        _dest_op_platforms = []
        for op in self.dest_ops:
            if op is not None:
                _dest_op_platforms.append(op.platform)
            else:
                _dest_op_platforms.append(TargetPlatform.FP32)
        return _dest_op_platforms

    @property
    def source_op_config(self) -> TensorQuantizationConfig:
        if self.source_op is not None:
            if isinstance(self.source_op, QuantableOperation):
                return self.source_op.config.output_quantization_config[self.src_idx]
            else:
                return None
        return None

    @property
    def source_op_platform(self) -> TargetPlatform:
        if self.source_op is None:
            return TargetPlatform.FP32
        else:
            return self.source_op.platform

    def copy(self, copy_value: bool = False):
        clone = QuantableVariable(super().copy(copy_value))
        if copy_value and self._fp32_value is not None:
            clone._fp32_value = self._fp32_value.clone()
        else:
            clone._fp32_value = self._fp32_value
        return clone


class DeviceSwitchOP(Operation):
    """This Class has been removed since PPQ 0.6.6"""

    def __init__(self, name: str, inputs: List[Variable] = None, outputs: List[Variable] = None) -> None:
        super().__init__(
            attributes={},
            name=name,
            op_type="PPQDeviceSwitch",
            platform=TargetPlatform.UNSPECIFIED,
            inputs=inputs,
            outputs=outputs,
        )


class QuantableGraph(GraphCommandProcessor):
    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.QUANTIZE_OPERATION:
            assert isinstance(command, QuantizeOperationCommand)
            return self.quantize_operation(command.op_name, command.target_platform, command.config)

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.QUANTIZE_OPERATION,
        ]

    def quantize_operation(
        self, operation_name: str, target_platform: TargetPlatform, quantization_config: OperationQuantizationConfig
    ) -> QuantableOperation:
        if operation_name not in self.graph.operations:
            raise KeyError(f"Operation {operation_name} is not in your graph, Please check your input.")

        operation = self._graph.operations[operation_name]
        quantized_operation = QuantableOperation(
            convert_from=operation,
            quantize_config=quantization_config,
            platform=target_platform,
        )

        # calling other chain responder to replace operation with quantized one.
        if self._next_command_processor is None:
            raise RuntimeError("To replace a operation, your processor chain must have a GraphReplacer Processor.")
        self._next_command_processor(ReplaceOperationCommand(operation_name, quantized_operation))

        # replace all related variable with quantable one.
        for var in quantized_operation.inputs + quantized_operation.outputs:
            if isinstance(var, QuantableVariable):
                continue
            self._next_command_processor(
                ReplaceVariableCommand(var_name=var.name, replace_to=QuantableVariable(convert_from=var))
            )
        quantized_operation.store_parameter_value()

    def dequantize_operation(self, operation_name: str) -> Operation:
        if operation_name not in self.graph.operations:
            raise KeyError(f"Operation {operation_name} is not in your graph, Please check your input.")
        operation = self._graph.operations[operation_name]
        if not isinstance(operation, QuantableOperation):
            return operation
        else:
            return operation.dequantize()

    def dequantize_graph(self, expire_device: str = "cpu"):
        """一个方便懒人的函数."""
        for operation in self.graph.operations.values():
            if isinstance(operation, QuantableOperation):
                operation.dequantize(expire_device=expire_device)

    def restore_quantize_state(self, expire_device: str = "cpu"):
        """一个方便懒人的函数."""
        for operation in self.graph.operations.values():
            if isinstance(operation, QuantableOperation):
                operation.restore_quantize_state(expire_device=expire_device)



import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern


class DynamicQuantizeMatMulIntegerPatternMatcher(PatternMatcher):
    def __init__(self, priority):

        pattern = Pattern(
            """
            input                   input               0 1 lhs_dyn_quant_0
            DynamicQuantizeLinear   lhs_dyn_quant_0     1 3 input matmul_0 mul_0 matmul_0
            MatMulInteger           matmul_0            4 1 lhs_dyn_quant_0 ? lhs_dyn_quant_0 ? cast_0
            Cast                    cast_0              1 1 matmul_0 mul_1
            Mul                     mul_0               2 1 lhs_dyn_quant_0 ? mul_1
            Mul                     mul_1               2 1 cast_0 mul_0 output
            output                  output              1 0 mul_1
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionDynamicQuantizeMatMulInteger'."""
        return "FusionDynamicQuantizeMatMulInteger"

    def rewrite(self, opset=11):
        lhs_dyn_quant_0 = self.lhs_dyn_quant_0
        matmul_0 = self.matmul_0
        mul_0 = self.mul_0

        if len(lhs_dyn_quant_0.outputs[0].outputs) == 1:
            input_variable = lhs_dyn_quant_0.inputs[0]
            input_variable.outputs.remove(lhs_dyn_quant_0)

            weight_variable = matmul_0.inputs[1]
            weight_scale_variable = mul_0.inputs[1]
            weight_zp_variable = matmul_0.inputs[3]

            weight_variable.outputs.remove(matmul_0)
            weight_scale_variable.outputs.remove(mul_0)
            weight_zp_variable.outputs.remove(matmul_0)

            output_variable = self.mul_1.outputs[0]
            output_variable.inputs.clear()
            return {
                matmul_0.name: {
                    "op": "DynamicQuantizeMatMul",
                    "inputs": [input_variable, weight_variable, weight_scale_variable, weight_zp_variable],
                    "outputs": [output_variable],
                    "domain": "com.microsoft",
                    "name": matmul_0.name
                }
            }
        else:
            return {}

class DynamicQuantizeMatMulPatternMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input                   input               0 1 lhs_dyn_quant_0
            DynamicQuantizeLinear   lhs_dyn_quant_0     1 3 input lhs_dq_0 lhs_dq_0 lhs_dq_0
            DequantizeLinear        lhs_dq_0            3 1 lhs_dyn_quant_0 lhs_dyn_quant_0 lhs_dyn_quant_0 matmul_0
            DequantizeLinear        rhs_dq_0            3 1 ? ? ? matmul_0
            MatMul                  matmul_0            2 1 lhs_dq_0 rhs_dq_0 output
            output                  output              1 0 matmul_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionDynamicQuantizeMatMul'."""
        return "FusionDynamicQuantizeMatMul"

    def rewrite(self, opset=11):
        lhs_dyn_quant_0 = self.lhs_dyn_quant_0
        rhs_dq_0 = self.rhs_dq_0
        matmul_0 = self.matmul_0

        if len(lhs_dyn_quant_0.outputs[0].outputs) == 1:
            input_variable = lhs_dyn_quant_0.inputs[0]
            input_variable.outputs.remove(lhs_dyn_quant_0)

            weight_variable = rhs_dq_0.inputs[0]
            weight_scale_variable = rhs_dq_0.inputs[1]
            weight_zp_variable = rhs_dq_0.inputs[2]

            weight_variable.outputs.remove(rhs_dq_0)
            weight_scale_variable.outputs.remove(rhs_dq_0)
            weight_zp_variable.outputs.remove(rhs_dq_0)

            output_variable = matmul_0.outputs[0]
            output_variable.inputs.clear()
            return {
                matmul_0.name: {
                    "op": "DynamicQuantizeMatMul",
                    "inputs": [input_variable, weight_variable, weight_scale_variable, weight_zp_variable],
                    "outputs": [output_variable],
                    "domain": "com.microsoft",
                    "name": matmul_0.name
                }
            }
        else:
            return {}

class DynamicQuantizeMatMulBiasRHSPatternMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input                   input               0 1 dyn_matmul_0
            DynamicQuantizeMatMul   dyn_matmul_0        4 1 input ? ? ? bias_add_0
            Add                     bias_add_0          2 1 dyn_matmul_0 ? output
            output                  output              1 0 bias_add_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionDynamicQuantizeMatMulBiasRHS'."""
        return "FusionDynamicQuantizeMatMulBiasRHS"

    def rewrite(self, opset=11):
        dyn_matmul_0 = self.dyn_matmul_0
        bias_add_0 = self.bias_add_0

        if isinstance(bias_add_0.inputs[1], osg.Constant) and \
            bias_add_0.inputs[1].values is not None and \
            len(bias_add_0.inputs[1].values.shape) <= 1:
            input_variable = dyn_matmul_0.inputs[0]
            weight_variable = dyn_matmul_0.inputs[1]
            weight_scale_variable = dyn_matmul_0.inputs[2]
            weight_zp_variable = dyn_matmul_0.inputs[3]

            input_variable.outputs.remove(dyn_matmul_0)
            weight_variable.outputs.remove(dyn_matmul_0)
            weight_scale_variable.outputs.remove(dyn_matmul_0)
            weight_zp_variable.outputs.remove(dyn_matmul_0)

            bias_variable = bias_add_0.inputs[1]
            bias_variable.outputs.remove(bias_add_0)

            output_variable = bias_add_0.outputs[0]
            output_variable.inputs.clear()

            return {
                dyn_matmul_0.name: {
                    "op": "DynamicQuantizeMatMul",
                    "inputs": [input_variable, weight_variable, weight_scale_variable, weight_zp_variable, bias_variable],
                    "outputs": [output_variable],
                    "domain": "com.microsoft",
                    "name": dyn_matmul_0.name
                }
            }
        else:
            return {}

class DynamicQuantizeMatMulBiasLHSPatternMatcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input                   input               0 1 dyn_matmul_0
            DynamicQuantizeMatMul   dyn_matmul_0        4 1 input ? ? ? bias_add_0
            Add                     bias_add_0          2 1 ? dyn_matmul_0 output
            output                  output              1 0 bias_add_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionDynamicQuantizeMatMulBiasLHS'."""
        return "FusionDynamicQuantizeMatMulBiasLHS"

    def rewrite(self, opset=11):
        dyn_matmul_0 = self.dyn_matmul_0
        bias_add_0 = self.bias_add_0

        if isinstance(bias_add_0.inputs[0], osg.Constant) and \
            bias_add_0.inputs[0].values is not None and \
            len(bias_add_0.inputs[0].values.shape) <= 1:
            input_variable = dyn_matmul_0.inputs[0]
            weight_variable = dyn_matmul_0.inputs[1]
            weight_scale_variable = dyn_matmul_0.inputs[2]
            weight_zp_variable = dyn_matmul_0.inputs[3]

            input_variable.outputs.remove(dyn_matmul_0)
            weight_variable.outputs.remove(dyn_matmul_0)
            weight_scale_variable.outputs.remove(dyn_matmul_0)
            weight_zp_variable.outputs.remove(dyn_matmul_0)

            bias_variable = bias_add_0.inputs[0]
            bias_variable.outputs.remove(bias_add_0)

            output_variable = bias_add_0.outputs[0]
            output_variable.inputs.clear()

            return {
                dyn_matmul_0.name: {
                    "op": "DynamicQuantizeMatMul",
                    "inputs": [input_variable, weight_variable, weight_scale_variable, weight_zp_variable, bias_variable],
                    "outputs": [output_variable],
                    "domain": "com.microsoft",
                    "name": dyn_matmul_0.name
                }
            }
        else:
            return {}

register_fusion_pattern(DynamicQuantizeMatMulIntegerPatternMatcher(1))
#register_fusion_pattern(DynamicQuantizeMatMulPatternMatcher(1))
register_fusion_pattern(DynamicQuantizeMatMulBiasRHSPatternMatcher(1))
register_fusion_pattern(DynamicQuantizeMatMulBiasLHSPatternMatcher(1))



# PPQ 全局配置，你可以自由修改下列属性以达成特定目的。
# PPQ System configuration
# You can modify following codes for your own purpose.

MAX_RECURSION_DEPTH = 5000
import sys
sys.setrecursionlimit(MAX_RECURSION_DEPTH)

# Observer 中，最小 scale 限制，所有小于该值的 scale 将被该值覆盖
OBSERVER_MIN_SCALE = 1e-8
# Observer 中，最小 scale 的手动覆盖属性
OBSERVER_MIN_SCALE_MANUL_OVERRIDE = 'OBSERVER_MIN_SCALE_MANUL_OVERRIDE'
# 当出现 scale < min_scale 的情况时，是否给出警报 
OBSERVER_WARNING = False
# Observer 中 kl 散度的计算设备
OBSERVER_KL_COMPUTING_DEVICE = 'cpu'
# Observer 中 hist 箱子的个数
OBSERVER_KL_HIST_BINS = 4096
# Observer 中 hist 箱子的属性
OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE = 'OBSERVER_KL_HIST_BINS_MANUL_OVERRIDE'
# Observer 中 percentile 的参数
OBSERVER_PERCENTILE = 0.9999
# Observer 中 percentile 参数的属性
OBSERVER_PERCENTILE_MANUL_OVERRIDE = 'OBSERVER_PERCENTILE_MANUL_OVERRIDE'
# Observer 中 mse 校准方法 hist 箱子的个数
OBSERVER_MSE_HIST_BINS = 2048
# Observer 中 mse 计算的间隔，间隔越小，所需时间越长
OBSERVER_MSE_COMPUTE_INTERVAL = 8
# Floating MSE Observer 的样本个数
OBSERVER_FLOATING_MSE_FETCHES = 4096
# Isotone Observer 的监听轴
# 对于 Softmax 激活函数而言，Isotone Observer 的监听轴应该与 Softmax 操作所规约的轴相同
# 对于 Sigmoid 激活函数而言，Isotone Observer 的监听轴应该设置为 Batch 所在的轴
OBSERVER_ISOTONE_OBSERVER_AXIS = 'OBSERVER_ISOTONE_OBSERVER_AXIS'

# 读取 Onnx 图时，将图中所有以 Constant 节点作为输入的变量转换为 Parameter Variable
FORMATTER_FORMAT_CONSTANT_INPUT = True
# 读取 Onnx 图时，合并图中的 Bias add 节点(Conv, ConvTranspose, Gemm)
FORMATTER_FUSE_BIAS_ADD = True
# 读取 Onnx 图时，合并图中的 Batchnorm 节点(Conv, ConvTranspose, Gemm)
FORMATTER_FUSE_BN = True
# 读取 Onnx 图时，将单独的 Batchnorm 替换为卷积
FORMATTER_REPLACE_BN_TO_CONV = True
# 读取 Onnx 图时，移除图中的 Identity 节点
FORMATTER_REMOVE_IDENTITY = True
# 读取 Onnx 图时，移除图中的孤立节点
FORMATTER_REMOVE_ISOLATED = True

# PASSIVE OPERATIONS 是那些不参与计算的 Op, 这些 op 的输入与输出将直接共享 scale
# 同时这些 op 前后的定点过程将被直接停用
PASSIVE_OPERATIONS = {
    'MaxPool', 'GlobalMaxPool', 'Reshape', 'Flatten', 'Identity', 'Dropout'
    'Slice', 'Pad', 'Split', 'Transpose', 'Interp', 'Squeeze', 'Unsqueeze'}
# COPUTING OP 是所有计算层，该属性被用于联合定点和子图切分
COMPUTING_OP = {'Conv', 'Gemm', 'ConvTranspose', 'MatMul', 'Attention', 'PPQBiasFusedMatMul'}
# SOI OP 是所有产生 SOI 输出的节点类型，该属性被用于子图切分
SOI_OP = {'TopK', 'Shape', 'NonMaxSuppression'}

# 强制联合定点的算子种类
TYPES_FOR_ALIGNMENT = {
    'Elementwise': {'Add', 'Sub', 'Sum'}, 
    'Concat': {'Concat'}, 
    'Pooling': {'AveragePool', 'GlobalAveragePool'}}
# 强制联合定点手动覆盖
ALIGNMENT_MANUL_OVERRIDE = 'ALIGNMENT_MANUL_OVERRIDE'

# Training Based Pass 抽样大小
NUM_OF_CHECKPOINT_FETCHS = 4096

# SUB GRAPH 最大深度
# PPQ 使用子图切割算法寻找子图，这个参数控制了子图的大小。
OPTIM_ADVOPT_GRAPH_MAXDEPTH = 4
# ROUNDING LOSS 系数
OPTIM_ADVOPT_RLOSS_MULTIPLIER = 1
# 是否使用 LR SCEHDULER
OPTIM_ADVOPT_USING_SCEHDULER = True

# ONNX 导出图的时候，opset的版本，这玩意改了可能就要起飞了
ONNX_EXPORT_OPSET = 11
# ONNX 导出图的时候，onnx version，这玩意改了可能就要起飞了
ONNX_VERSION = 6

ONNX_DOMAIN = 'ai.onnx'
CAFFE_DOMAIN = 'ppq.caffe'
DEFAULT_OPSET_DOMAIN  = 'ai.onnx'
DEFAULT_OPSET_VERSION = 11
STRICT_OPSET_CHECKING = False

# LSTM 算子的权重缓存属性
LSTM_FLATTEN_WEIGHT_ATTRIB = 'LSTM_FLATTEN_WEIGHT_ATTRIB'
# GRU 算子的权重缓存属性
GRU_FLATTEN_WEIGHT_ATTRIB  = 'GRU_FLATTEN_WEIGHT_ATTRIB'
# 一个属性标记计算图是否已经被调度
IS_DISPATCHED_GRAPH = 'IS_DISPATCHED_GRAPH'
# 图上用于表示 Opset 的属性
GRAPH_OPSET_ATTRIB = 'GRAPH_OPSET'

# -------------- 以下属性已经过时：

# PPLCUDA 中所有需要与 Conv 融合的激活函数
PPLCUDA_ACTIVATIONS = {'Clip', 'LeakyRelu', 'Relu', 'Sigmoid'}
ORT_OOS_FUSE_START_OPS = {'Conv', 'GlobalAveragePool', 'AveragePool', 'Add', 'Mul', 'MatMul'}
# LINEAR ACTIVATIONS 是所有线性激活层，PPQ 将执行计算层与线性激活层的联合定点，不论后端是否真的做了图融合。
# 事实上就算后端不融合这些层，执行联合定点也是有益无害的。
LINEAR_ACTIVATIONS = {'Relu', 'Clip'}
COMPELING_OP_TYPES = {'Add',' Sub', 'Concat'}

# 误差容忍度
CHECKPOINT_TOLERANCE = 1

# 要做 Bias Correction 的算子种类
BIAS_CORRECTION_INTERST_TYPE = {'Conv', 'Gemm', 'ConvTranspose'}

# 导出 qdq 节点时是否需要导出状态已经是 overlap 的节点
EXPORT_OVERLAPPED_CONFIG = False



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import argparse

from xslim.logger import logger

from . import quantize_onnx_model

parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument("-c", "--config", required=False, default=None, help="Path to the Xquant Config.")
parser.add_argument("-i", "--input_path", required=False, default=None, help="Path to the Input ONNX Model.")
parser.add_argument("-o", "--output_path", required=False, default=None, help="Path to the Output ONNX Model.")
parser.add_argument("--fp16", required=False, action="store_true", help="convert onnx model to fp16.")
parser.add_argument("--dynq", required=False, action="store_true", help="convert onnx model to dynq.")
parser.add_argument("--ignore_op_types", required=False, default="", help="Ignore op types.")
parser.add_argument("--ignore_op_names", required=False, default="", help="Ignore op names.")

if __name__ == "__main__":
    args = parser.parse_args()

    if args.config is None and (args.input_path is None or args.output_path is None):
        parser.print_help()
        exit(1)

    if args.config is None:
        precesion_level = 100

        if args.fp16:
            precesion_level = 4
        elif args.dynq:
            precesion_level = 3

        if precesion_level == 3:
            logger.info("No config provided, using default config, dynamic quantization...")
        elif precesion_level == 4:
            logger.info("No config provided, using default config, convert onnx model to fp16...")
        elif precesion_level >= 100:
            logger.info("No config provided, using default config, only simplify onnx model...")

        args.config = {
            "quantization_parameters": {
                "precision_level": precesion_level,
                "ignore_op_types": [i for i in args.ignore_op_types.split(",") if i != ""],
                "ignore_op_names": [i for i in args.ignore_op_names.split(",") if i != ""],
            },
        }

    quantize_onnx_model(args.config, args.input_path, args.output_path)



from abc import ABCMeta
from typing import Dict, List

from ...core import QuantizationStates, TensorQuantizationConfig, ppq_debug_function
from ...executor import QuantOPRuntimeHook
from ...IR import QuantableOperation, Variable

from .base import BaseTensorObserver
from .range import TorchMinMaxObserver, TorchPercentileObserver

OBSERVER_TABLE = {
    "minmax": TorchMinMaxObserver,
    "percentile": TorchPercentileObserver,
}


class TensorObserverFactroy:
    def __init__(self) -> None:
        raise NotImplementedError(
            "Observer Factory can not be initialized, use TensorObserverFactroy.build_observer instead."
        )

    @classmethod
    def build_observer(cls, variable: Variable, config: TensorQuantizationConfig) -> BaseTensorObserver:
        algorithm = str(config.observer_algorithm.lower())
        if algorithm not in OBSERVER_TABLE:
            raise ValueError(
                f"Observer type not understand, Except one of {OBSERVER_TABLE.keys()}, "
                f"while {str(algorithm)} was given."
            )
        return OBSERVER_TABLE[algorithm](watch_on=variable, quant_cfg=config)


class CalibrationHook(QuantOPRuntimeHook):
    def __init__(self, operation: QuantableOperation, observer_table: Dict[Variable, BaseTensorObserver]) -> None:
        self._operation = operation
        self._observer_table = observer_table
        super().__init__(operation)

    def pre_forward_hook(self, inputs: list, quant_inputs: list, quant_configs: List[TensorQuantizationConfig]) -> list:
        for input_var, quant_config in zip(inputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(input_var)
        return quant_inputs

    def post_forward_hook(
        self, outputs: list, quant_outputs: list, quant_configs: List[TensorQuantizationConfig]
    ) -> list:
        for output_var, quant_config in zip(outputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(output_var)
        return quant_outputs

    def render_quantization_config(self):
        for _, observer in self._observer_table.items():
            observer.render_quantization_config()
            observer.report()

    def __str__(self) -> str:
        return "".join([observer.__str__() + "\n" for _, observer in self._observer_table.items()])


class OperationObserver(metaclass=ABCMeta):
    def __init__(
        self,
        operation: QuantableOperation,
        monitor_parameter: bool = True,
        monitor_outputs: bool = True,
        monitor_inputs: bool = True,
    ) -> None:
        if not isinstance(operation, QuantableOperation):
            raise TypeError(f"Only QuantableOP instance can apply an Observer, " f"while {type(operation)} was given.")

        self._operation = operation
        self._hook = self.build_hook(
            monitor_parameter=monitor_parameter, monitor_outputs=monitor_outputs, monitor_inputs=monitor_inputs
        )

    def render_quantization_config(self):
        self.hook.render_quantization_config()

    def build_hook(self, monitor_parameter: bool, monitor_outputs: bool, monitor_inputs: bool) -> CalibrationHook:
        assert isinstance(self._operation, QuantableOperation)
        observer_table = {}
        for var, config in zip(self._operation.inputs, self._operation.config.input_quantization_config):
            if config.state == QuantizationStates.INITIAL:
                if var in self._operation.parameters and monitor_parameter:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)
                elif monitor_inputs:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)

        if monitor_outputs:
            for var, config in zip(self._operation.outputs, self._operation.config.output_quantization_config):
                if config.state == QuantizationStates.INITIAL:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)

        return CalibrationHook(operation=self._operation, observer_table=observer_table)

    @property
    def hook(self) -> CalibrationHook:
        return self._hook

    @ppq_debug_function
    def report(self) -> str:
        return str(self._hook)



"""PPQ Core Data Structure Abstraction PPQ 核心数据结构抽象.

You are not allowed to modify this 请勿修改此文件
"""
from enum import Enum
from typing import Any, List, Union

import numpy as np
import torch
from numpy import dtype as np_type
from numpy import ndarray
from torch import Tensor
from torch import dtype as torch_type


class DataType(Enum):
    """
        DataType defines all PPQ internal data type and its enumeration value.
            ATTENTION: PPQ shares same data type enumeration value with Onnx.

        System maintainers and modifier are supposed to keep this corresponding.
        Cause OnnxExporter directly use this value to export PPQ graph towards Onnx.
    """
    INT4   = -1 # Onnx doesn't have this definition
    UINT4  = -2 # Onnx doesn't have this definition
    INT8   = 3  # Onnx.TensorProto.DataType.INT8
    UINT8  = 2  # Onnx.TensorProto.DataType.UINT8
    INT16  = 5  # Onnx.TensorProto.DataType.INT16
    UINT16 = 4  # Onnx.TensorProto.DataType.UINT16
    INT32  = 6  # Onnx.TensorProto.DataType.INT32
    UINT32 = 12 # Onnx.TensorProto.DataType.UINT32
    INT64  = 7  # Onnx.TensorProto.DataType.INT64
    UINT64 = 13 # Onnx.TensorProto.DataType.UINT64

    FP16 = 10 # Onnx.TensorProto.DataType.FLOAT16
    FP32 = 1  # Onnx.TensorProto.DataType.FLOAT
    FP64 = 11 # Onnx.TensorProto.DataType.DOUBLE

    BOOL = 9  # Onnx.TensorProto.DataType.BOOL
    COMPLEX128 = 15 # Onnx.TensorProto.DataType.COMPLEX128
    COMPLEX64 = 14 # Onnx.TensorProto.DataType.COMPLEX64
    NONETYPE  = 0 # Onnx.TensorProto.DataType.UNSPECIFIED

    @ classmethod
    def convert_from_numpy(cls, dtype: np_type):
        numpy_converting_dict = {
            np_type('bool'):    DataType.BOOL,
            np_type('uint8'):   DataType.UINT8,
            np_type('int8'):    DataType.INT8,
            np_type('int16'):   DataType.INT16,
            np_type('int32'):   DataType.INT32,
            np_type('int64'):   DataType.INT64,
            np_type('float16'): DataType.FP16,
            np_type('float32'): DataType.FP32,
            np_type('float64'): DataType.FP64,
        }
        if dtype not in numpy_converting_dict:
            raise TypeError(f'Numpy type {dtype} is not included in ppq now. '
                'please contact with system developer.')
        else:
            return numpy_converting_dict[dtype]

    @ classmethod
    def convert_from_torch(cls, dtype: torch_type):
        torch_converting_dict = {
            torch.bool:    DataType.BOOL,
            torch.uint8:   DataType.UINT8,
            torch.int8:    DataType.INT8,
            torch.int16:   DataType.INT16,
            torch.int32:   DataType.INT32,
            torch.int64:   DataType.INT64,
            torch.float16: DataType.FP16,
            torch.float32: DataType.FP32,
            torch.float64: DataType.FP64,
        }
        if dtype not in torch_converting_dict:
            raise TypeError(f'Torch dtype {dtype} is not included in ppq now. '
                'please contact with system developer.')
        else:
            return torch_converting_dict[dtype]

    @ classmethod
    def to_numpy(cls, dtype) -> np_type:
        numpy_converting_dict = {
            DataType.BOOL:  np_type('bool'),
            DataType.UINT8: np_type('uint8'),
            DataType.INT8:  np_type('int8'),
            DataType.INT16: np_type('int16'),
            DataType.INT32: np_type('int32'),
            DataType.INT64: np_type('int64'),
            DataType.FP16:  np_type('float16'),
            DataType.FP32:  np_type('float32'),
            DataType.FP64:  np_type('float64'),
        }
        assert isinstance(dtype, DataType)
        return numpy_converting_dict[dtype]

    @ classmethod
    def to_torch(cls, dtype) -> torch_type:
        torch_converting_dict = {
            DataType.BOOL:  torch.bool,
            DataType.UINT8: torch.uint8,
            DataType.INT8:  torch.int8,
            DataType.INT16: torch.int16,
            DataType.INT32: torch.int32,
            DataType.INT64: torch.int64,
            DataType.FP16:  torch.float16,
            DataType.FP32:  torch.float32,
            DataType.FP64:  torch.float64,
        }
        assert isinstance(dtype, DataType)
        return torch_converting_dict[dtype]


class TensorMeta:
    def __init__(
        self, dtype: DataType, shape: List[int],
        tensor_name: str = None) -> None:
        """TensorMeta structure described metadata of a tensor.

        Which includes tensor's data type and shape.
        TensorMeta is necessary to initialize quantization configuration and hooks,
        and is needed to compute the number of input channels.
        Args:
            dtype (DataType):
                A DataType enumeration described tensor type.
            shape (List[int]):
                A int list contains size of each dimensions.
            tensor_name (str, optional): Not yet used.
        """
        if not isinstance(dtype, DataType):
            raise TypeError(f'Can not create Tensor Meta with dtype {type(dtype)}, '
                            'only ppq.core.DataType instance is acceptable here.')
        self.dtype = dtype
        self.shape = shape
        self.name  = tensor_name

    @ classmethod
    def parsing_from_numpy_ndarray(cls, numpy_array: ndarray, name: str = None):
        shape = list(numpy_array.shape)
        dtype = DataType.convert_from_numpy(numpy_array.dtype)
        return TensorMeta(dtype=dtype, shape=shape,tensor_name=name)

    @ classmethod
    def parsing_from_torch_tensor(cls, torch_tensor: Tensor, name: str = None):
        if not isinstance(torch_tensor, Tensor):
            raise TypeError(f'Can not parse meta data for {type(torch_tensor)} instance, '
                'it should be torch.Tensor object.')
        shape = list(torch_tensor.shape)

        # for tensor scalar, which do not have an valid shape
        # just mannully give a empty list to them.
        if not shape: shape = []

        dtype = DataType.convert_from_torch(torch_tensor.dtype)
        return TensorMeta(dtype=dtype, shape=shape, tensor_name=name)

    def create_tensor(self, device: str, fill_value: Any = 0):
        return torch.Tensor(size=self.shape, device='cpu').fill_(
            fill_value).type(dtype=DataType.to_torch(self.dtype)).to(device)

    def create_ndarray(self, fill_value: Any = 0):
        return ndarray(shape=self.shape,
            dtype=DataType.to_numpy(self.dtype)).fill(fill_value)

    def __str__(self) -> str:
        return f'Tensor({self.name}) meta: dtype({self.dtype}), shape({self.shape})'
    
    def copy(self):
        if self.shape is not None:
            return TensorMeta(dtype=self.dtype, shape=self.shape.copy(), tensor_name=self.name)
        else: return TensorMeta(dtype=self.dtype, shape=None, tensor_name=self.name)


class OperationMeta:
    def __init__(self,
        input_metas: List[TensorMeta], output_metas: List[TensorMeta],
        operation_name: str, operation_type: str, executing_order: int) -> None:
        """OperationMeta structure describes all related tensor metadata of an
        operation.

        It naturally is a collection of TensorMeta.
        Take a look at TensorMeta to get more information.
        Args:
            input_metas (List[TensorMeta]):
                A collection contains all input tensors' metadata.
                ATTENTION: All parameters are considered as input in PPQ.
            output_metas (List[TensorMeta]):
                A collection contains all output tensors' metadata.
            operation_name (str): Not yet used.
            operation_type (str): Not yet used.
            executing_order (int): a int value represents the executing order of this operation.
                (order 0 means this operation is the first operation to be executed)
        """
        assert isinstance(input_metas, list), 'can only accept list object here.'
        assert isinstance(output_metas, list), 'can only accept list object here.'

        self.input_metas  = input_metas
        self.output_metas = output_metas

        self.operation_name = operation_name
        self.operation_type = operation_type
        self.executing_order = executing_order

    def __str__(self) -> str:
        return 'Inputs: '.join(str([_ for _ in self.input_metas])) + \
            'Outputs: '.join(str([_ for _ in self.input_metas]))

    @ property
    def num_of_input(self):
        return len(self.input_metas)

    @ property
    def num_of_output(self):
        return len(self.output_metas)

    def copy(self):
        return OperationMeta(
            input_metas = [meta.copy() for meta in self.input_metas],
            output_metas = [meta.copy() for meta in self.output_metas],
            operation_name=self.operation_name, operation_type=self.operation_type, 
            executing_order=self.executing_order)


def convert_any_to_python_primary_type(
    x: Union[torch.Tensor, np.ndarray, int, float, list, str],
    accept_none: bool=True) -> Union[int, float, list, str]:
    if x is None and accept_none: return None
    if x is None and not accept_none: raise ValueError('Trying to convert an empty value.')
    if isinstance(x, list) or isinstance(x, tuple): return list(x)
    elif isinstance(x, int) or isinstance(x, float): return x
    elif isinstance(x, torch.Tensor):
        if x.numel() == 0 and accept_none: return None
        if x.numel() == 0 and not accept_none: raise ValueError('Trying to convert an empty value.')
        if str(x.device) != 'cpu': x = x.cpu()
        if x.numel() == 1: return x.item()
        if x.numel()  > 1: return x.tolist()
    elif isinstance(x, np.ndarray):
        if x.size == 0 and accept_none: return None
        if x.size == 0 and not accept_none: raise ValueError('Trying to convert an empty value.')
        if x.size == 1: return x.reshape((1, )).tolist()[0]
        if x.size  > 1: return x.tolist()
    elif isinstance(x, str):
        return x
    else:
        raise TypeError(f'input value {x}({type(x)}) can not be converted as python primary type.')


def convert_any_to_numpy(
    x: Union[torch.Tensor, np.ndarray, int, float, list, tuple],
    accept_none: bool=True) -> np.ndarray:
    if x is None and accept_none: return None
    if x is None and not accept_none: raise ValueError('Trying to convert an empty value.')
    if isinstance(x, np.ndarray): return x
    elif isinstance(x, int) or isinstance(x, float): return np.array([x, ])
    elif isinstance(x, torch.Tensor):
        if x.numel() == 0 and accept_none: return None
        if x.numel() == 0 and not accept_none: raise ValueError('Trying to convert an empty value.')
        if x.numel() >= 1: return x.detach().cpu().numpy()
    elif isinstance(x, list) or isinstance(x, tuple):
        return np.array(x)
    else:
        raise TypeError(f'input value {x}({type(x)}) can not be converted as numpy type.')


def convert_any_to_torch_tensor(
    x: Union[torch.Tensor, np.ndarray, int, float, list, tuple],
    accept_none: bool=True, dtype: torch.dtype=None, device='cpu') -> torch.Tensor:
    if x is None and accept_none: return None
    if x is None and not accept_none: raise ValueError('Trying to convert an empty value.')
    if isinstance(x, list) or isinstance(x, tuple):
        if all([type(element) == int for element in x]):
            if dtype is None: dtype=torch.int64
        return torch.tensor(x, dtype=dtype, device=device)
    elif isinstance(x, int):
        if dtype is None: dtype=torch.int64
        return torch.tensor(x, dtype=dtype, device=device)
    elif isinstance(x, float):
        if dtype is None: dtype=torch.float32
        return torch.tensor(x, dtype=dtype, device=device)
    elif isinstance(x, torch.Tensor):
        if dtype is not None: x = x.type(dtype)
        if device is not None: x = x.to(device)
        return x
    elif isinstance(x, np.ndarray):
        if dtype is None:
            dtype = DataType.convert_from_numpy(x.dtype)
            dtype = DataType.to_torch(dtype)
        return torch.tensor(x.copy(), dtype=dtype, device=device)
    else:
        raise TypeError(f'input value {x}({type(x)}) can not be converted as torch tensor.')


def convert_primary_type_to_list(
    x:Union[int, float, list, tuple]) -> list:
    if isinstance(x, list) or isinstance(x, tuple): return list(x)
    elif isinstance(x, int) or isinstance(x, float): return [x]
    else:
        raise TypeError(f'input value {x}({type(x)}) can not be converted as python list.')


def convert_any_to_string(x: Union[torch.Tensor, np.ndarray, int, float, list, tuple]) -> str:
    if isinstance(x, int) or isinstance(x, float):
        return x.__str__()
    elif isinstance(x, list) or isinstance(x, tuple):
        return '[{content}]'.format(content=''.join([convert_any_to_string(_) + ',' for _ in x]))
    elif isinstance(x, torch.Tensor):
        return convert_any_to_string(convert_any_to_numpy(x))
    elif isinstance(x, np.ndarray):
        return str(x.tobytes())
    else: return x.__str__()



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import logging
import os
from enum import Enum

from .logger import (xslim_debug, xslim_error, xslim_info, xslim_trace,
                     xslim_warning)


def _get_version():
    # Prefer the version from the local source tree (VERSION_NUMBER) when available.
    try:
        version_file = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            'VERSION_NUMBER'
        )
        with open(version_file, encoding='utf-8') as f:
            return f.read().strip()
    except (FileNotFoundError, OSError):
        pass
    # Fall back to the installed package metadata when VERSION_NUMBER is not present.
    try:
        from importlib.metadata import version, PackageNotFoundError
        return version('xslim')
    except (ImportError, PackageNotFoundError):
        return "unknown"


class XQUANT_GLOBAL_CONFIGURATION:
    def __init__(self) -> None:
        import torch

        self.cuda_support = torch.cuda.is_available()

        self.min_block_size = 10

        self.max_block_size = 20

        self.merge_block_step = 4

        self.fine_tune_epoch = 2

        self.equalization_iterations = 10

        self.max_bits = 12

        self.analyse_steps = 16

        self.version = _get_version()


PASSIVE_OPERATIONS = {
    "MaxPool",
    "GlobalMaxPool",
    "Reshape",
    "Flatten",
    "Identity",
    "Dropout",
    "Slice",
    "Pad",
    "Split",
    "Transpose",
    "Interp",
    "Squeeze",
    "Unsqueeze",
    "Gather",
}

COMPUTING_OP = {"Conv", "Gemm", "ConvTranspose", "MatMul", "BatchMatMul"}

BIAS_CORRECTION_INTERST_TYPE = {
    "Conv",
    "Gemm",
    "ConvTranspose",
    # "LayerNormalization",
    # "InstanceNormalization",
    # "GroupNormalization",
}

OBSERVER_FLOATING_MSE_FETCHES = 4096

OBSERVER_MIN_SCALE_THRESHOLD = 2**-23

OBSERVER_MAX_SCALE_THRESHOLD = 2**8

OBSERVER_MAX_BIAS_VAL = 2**20

OBSERVER_PERCENTILE = 0.9999

OBSERVER_SIGMOID_MAX_VALUE = 10

MIN_ONNX_OPSET_VERSION = 17

GLOBAL_FUNCTIONS_MAPPING = "GLOBAL_FUNCTIONS_MAPPING"


class AutoFinetuneLevel(Enum):
    DO_NOTHING = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3


class PrecisionLevel(Enum):
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3
    LEVEL_4 = 4
    LEVEL_None = 100


XQUANT_CONFIG = XQUANT_GLOBAL_CONFIGURATION()



from math import ceil
from typing import Callable, Dict, Iterable, List, Tuple

import torch
import torch.nn.functional as F
from tqdm import tqdm
from xslim.logger import logger

from ...core import QuantizationProperty, QuantizationStates
from ...executor import BaseGraphExecutor
from ...executor.base import OPERATION_FORWARD_TABLE
from ...IR import BaseGraph, Operation, QuantableOperation, Variable
from ...IR.search import Path, SearchableGraph, TraversalCommand
from ..measure import torch_mean_square_error
from ..observer import CalibrationHook, OperationObserver
from ..observer.range import TorchHistObserver
from ..qfunction import BaseQuantFunction, PPQLinearQuantFunction
from .base import QuantizationOptimizationPass

OPTIMIZATION_LAYERTYPE_CONFIG = {
    1: {
        "Relu",
        "MaxPool",
        "GlobalMaxPool",
        "PRelu",
        "AveragePool",
        "GlobalAveragePool",
    },  # support more relay operation types
}
EQUALIZATION_OPERATION_TYPE = {"Conv", "Gemm", "ConvTranspose"}  # support all computing op types


class SSDEqualizationPass(QuantizationOptimizationPass):
    """PPQ Custimized Layerwise Equalization Pass.

    This is another layerwise equalization pass which takes quantization error into consideration, for more
    details of equalization, please refer to LayerwiseEqualizationPass

    Equalization algo will only take effect when the estimated loss is less than loss_threshold * original_loss

    Compared with LayerwiseEqualizationPass, it guarantees a better result than the original model while consuming
    more time for loss estimation
    """

    def __init__(
        self,
        optimize_level: int = 1,
        channel_ratio: float = 0.5,
        loss_threshold: float = 0.8,
        layer_norm: bool = False,
        quant_func: BaseQuantFunction = PPQLinearQuantFunction,
        iteration: int = 3,
    ):
        """SSD Equalization Pass With Loss Checking.

        Args:
            optimize_level (int, optional): level of optimization. Only support level 1 for now.

            channel_ratio (float, optional): all values below this ratio of maximum value of corresponding weight will
            be clipped to this ratio when calculating scale. Defaults to 0.5.

            loss_threshold (float, optional): the estimated loss should be below this threshold * original_loss for algo
            to take effect. Defaults to 0.8.

            layer_norm (bool, optional): whether to apply weight normalization. Defaults to False.

            quant_func (BaseQuantFunction, optional): quantization function. Defaults to PPQLinearQuantFunction.

            iteration (int, optional): num of iterations to run, usually 3 would be enough. Defaults to 3.
        """
        self.channel_ratio = channel_ratio
        self.loss_threshold = loss_threshold
        self.layer_norm = layer_norm
        self.quant_func = quant_func
        self.start_op_types = EQUALIZATION_OPERATION_TYPE
        self.relay_op_types = OPTIMIZATION_LAYERTYPE_CONFIG[optimize_level]
        self.end_op_types = EQUALIZATION_OPERATION_TYPE
        self.iteration = iteration
        super().__init__(name="SSD Equalization Pass")

    def collect_all_pairs(self, graph: BaseGraph) -> List[List[Operation]]:
        def limitation(path: Path) -> bool:
            for op in path.tolist()[:-1]:
                if len(graph.get_downstream_operations(op)) != 1:
                    return False
            return True

        search_engine = SearchableGraph(graph)
        forward_matchings = search_engine.process(
            TraversalCommand(
                sp_expr=lambda x: x.type in self.start_op_types,
                rp_expr=lambda x, y: y.type in self.relay_op_types,
                ep_expr=lambda x: x.type in self.end_op_types,
                direction="down",
            )
        )
        return [path.tolist() for path in forward_matchings if limitation(path)]

    def collect_activation_range(
        self,
        pair: List[Operation],
        executor: BaseGraphExecutor,
        data_loader: Iterable,
        collate_fn: Callable,
        calib_steps: int,
    ) -> Dict[Operation, torch.Tensor]:
        """Collect activation ranges for Conv ops in the pair.

        Args:
            pair (List[Operation]): equalization pair
            executor (BaseGraphExecutor): graph executor
            data_loader (Iterable): data loader
            collate_fn (Callable): batch func
            calib_steps (int): calibration steps

        Returns:
            Dict[Operation, torch.Tensor]: activation range of Conv ops in the pair
        """
        op_act_ranges = {op: 0.0 for op in pair if op.type in self.start_op_types}
        output_names = [op.outputs[0].name for op in op_act_ranges]
        calib_step = 0
        for calib_epoch in range(ceil(calib_steps / len(data_loader))):
            for _, data in enumerate(data_loader):
                if collate_fn is not None:
                    data = collate_fn(data)
                outputs = executor.forward(inputs=data, output_names=output_names)
                for op in op_act_ranges:
                    op_conv_data = outputs[output_names.index(op.outputs[0].name)]
                    op_conv_data_relu = F.relu(op_conv_data)  # take abs value
                    op_conv_data_relu = op_conv_data_relu.permute(
                        1, 0, *(range(op_conv_data_relu.ndim)[2:])
                    ).contiguous()
                    op_conv_data_relu = op_conv_data_relu.reshape((op_conv_data_relu.shape[0], -1))
                    op_conv_data_relu = op_conv_data_relu.max(1)[0]
                    op_act_ranges[op] += op_conv_data_relu
                calib_step += 1
                if calib_step >= calib_steps:
                    break
        for op in op_act_ranges:
            op_act_ranges[op] /= calib_steps

        for op, op_range in op_act_ranges.items():
            assert isinstance(op_range, torch.Tensor)
            op_range_max = op_range.max()
            adjust_range = torch.where(
                op_range < op_range_max * self.channel_ratio, op_range_max * self.channel_ratio, op_range
            )
            op_act_ranges[op] = adjust_range
        return op_act_ranges

    def layer_weight_norm(self, pairs: List[List[Operation]]):
        for pair in pairs:
            op_first_weight = pair[0].parameters[0].value
            op_second_weight = pair[-1].parameters[0].value
            op_first_max = op_first_weight.abs().max()
            op_second_max = op_second_weight.abs().max()
            scale = (op_second_max / op_first_max).sqrt()
            pair[0].parameters[0].value = pair[0].parameters[0].value * scale
            if len(pair[0].parameters) > 1:
                pair[0].parameters[1].value = pair[0].parameters[1].value * scale
            pair[-1].parameters[0].value = pair[-1].parameters[0].value / scale

    def prepare_weight_for_equalization(self, pair: List[Operation]) -> Tuple[torch.Tensor]:
        first_computing_op_weight = pair[0].parameters[0].value
        last_computing_op_weight = pair[-1].parameters[0].value

        assert isinstance(first_computing_op_weight, torch.Tensor)
        assert isinstance(last_computing_op_weight, torch.Tensor)

        if pair[0].type == "Conv":
            # [C_out, C_in, K, K]
            C_out = first_computing_op_weight.shape[0]
            first_weight_range = first_computing_op_weight.reshape(C_out, -1).abs().max(dim=1)[0]

        elif pair[0].type == "Gemm":
            if pair[0].attributes.get("transB", 0):
                # [C_out, C_in]
                first_weight_range = first_computing_op_weight.abs().max(dim=1)[0]
            else:
                # [C_in, C_out]
                first_weight_range = first_computing_op_weight.abs().max(dim=0)[0]

        elif pair[0].type == "ConvTranspose":
            # [C_in, C_out // g, K, K]
            num_group = pair[0].attributes.get("group", 1)
            C_in, C_out_g, K1, K2 = first_computing_op_weight.shape
            first_computing_op_weight = first_computing_op_weight.reshape(num_group, C_in // num_group, C_out_g, K1, K2)
            first_computing_op_weight = first_computing_op_weight.permute(0, 2, 1, 3, 4).contiguous()
            first_computing_op_weight = first_computing_op_weight.reshape(num_group * C_out_g, -1)
            first_weight_range = first_computing_op_weight.abs().max(dim=1)[0]

        if pair[-1].type == "Conv":
            # [C_out, C_in // g, K, K]
            num_group = pair[-1].attributes.get("group", 1)
            C_out, C_in_g, K1, K2 = last_computing_op_weight.shape
            last_computing_op_weight = last_computing_op_weight.reshape(num_group, C_out // num_group, C_in_g, K1, K2)
            last_computing_op_weight = last_computing_op_weight.permute(0, 2, 1, 3, 4).contiguous()
            last_computing_op_weight = last_computing_op_weight.reshape(num_group * C_in_g, -1)
            last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]

        elif pair[-1].type == "Gemm":
            C_out = first_weight_range.shape[0]
            if pair[-1].attributes.get("transB", 0):
                if C_out != last_computing_op_weight.shape[1]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], C_out, -1
                    )
                    last_computing_op_weight = last_computing_op_weight.permute(1, 0, 2).contiguous().reshape(C_out, -1)
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
                else:
                    last_weight_range = last_computing_op_weight.abs().max(dim=0)[0]
            else:
                if C_out != last_computing_op_weight.shape[0]:
                    last_computing_op_weight = last_computing_op_weight.reshape(C_out, -1)
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]
                else:
                    last_weight_range = last_computing_op_weight.abs().max(dim=1)[0]

        elif pair[-1].type == "ConvTranspose":
            C_in = last_computing_op_weight.shape[0]
            last_weight_range = last_computing_op_weight.reshape(C_in, -1).abs().max(dim=1)[0]

        return first_weight_range, last_weight_range

    def write_back(self, pair: List[Operation], scale: torch.Tensor) -> None:
        first_computing_op_weight = pair[0].parameters[0].value
        last_computing_op_weight = pair[-1].parameters[0].value

        assert isinstance(first_computing_op_weight, torch.Tensor)
        assert isinstance(last_computing_op_weight, torch.Tensor)

        if pair[0].type == "Conv":
            pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(-1, 1, 1, 1)

        elif pair[0].type == "Gemm":
            if pair[0].attributes.get("transB", 0):
                pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(-1, 1)
            else:
                pair[0].parameters[0].value = first_computing_op_weight * scale.reshape(1, -1)

        elif pair[0].type == "ConvTranspose":
            num_group = pair[0].attributes.get("group", 1)
            C_in, C_out_g, K1, K2 = first_computing_op_weight.shape
            first_computing_op_weight = first_computing_op_weight.reshape(num_group, C_in // num_group, C_out_g, K1, K2)
            first_computing_op_weight = first_computing_op_weight * scale.reshape(num_group, 1, -1, 1, 1)
            pair[0].parameters[0].value = first_computing_op_weight.reshape(C_in, C_out_g, K1, K2)

        if len(pair[0].parameters) > 1:
            pair[0].parameters[1].value = pair[0].parameters[1].value * scale

        if pair[-1].type == "Conv":
            num_group = pair[-1].attributes.get("group", 1)
            C_out, C_in_g, K1, K2 = last_computing_op_weight.shape
            last_computing_op_weight = last_computing_op_weight.reshape(num_group, C_out // num_group, C_in_g, K1, K2)
            last_computing_op_weight = last_computing_op_weight / scale.reshape(num_group, 1, -1, 1, 1)
            pair[-1].parameters[0].value = last_computing_op_weight.reshape(C_out, C_in_g, K1, K2)

        elif pair[-1].type == "Gemm":
            if pair[-1].attributes.get("transB", 0):
                if scale.numel() != last_computing_op_weight.shape[1]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], scale.numel(), -1
                    )
                    last_computing_op_weight = last_computing_op_weight / scale.reshape(1, -1, 1)
                    pair[-1].parameters[0].value = last_computing_op_weight.reshape(
                        last_computing_op_weight.shape[0], -1
                    )
                else:
                    pair[-1].parameters[0].value = last_computing_op_weight / scale.reshape(1, -1)
            else:
                if scale.numel() != last_computing_op_weight.shape[0]:
                    last_computing_op_weight = last_computing_op_weight.reshape(
                        scale.numel(), -1, last_computing_op_weight.shape[-1]
                    )
                    last_computing_op_weight = last_computing_op_weight / scale.reshape(-1, 1, 1)
                    pair[-1].parameters[0].value = last_computing_op_weight.reshape(
                        -1, last_computing_op_weight.shape[-1]
                    )
                else:
                    pair[-1].parameters[0].value = last_computing_op_weight / scale.reshape(-1, 1)

        elif pair[-1].type == "ConvTranspose":
            pair[-1].parameters[0].value = last_computing_op_weight / scale.reshape(-1, 1, 1, 1)

    def one_step_equalization(
        self,
        pair: List[Operation],
        op_act_channel_range: Dict[Operation, torch.Tensor] = {},
        algo_type: int = 2,
        ssd_min_scale: float = 8,
        ssd_max_scale: float = 2,
        dfq_min_scale: float = 0.1,
        dfq_max_scale: float = 10,
        eps: float = 1e-8,
    ):
        """Equalization step with scale being calculated in the way specified
        by algo_type.

        Args:
            pair (List[Operation]): a list of operations representing a equalzation pair
            op_act_channel_range (Dict[Operation, torch.Tensor]): channel-wise activation range of all Conv ops in the graph
            algo_type (int, optional): minor algo type. 0 represents dfq algo, 1~3 represents ssd algo. Defaults to 2.
            ssd_min_scale (float, optional): minimum clip value of scale for ssd algo. Defaults to 8.
            ssd_max_scale (float, optional): maximum clip value of scale for ssd algo. Defaults to 2.
            dfq_min_scale (float, optional): minimum clip value of scale for dfq algo. Defaults to 0.1.
            dfq_max_scale (float, optional): maximum clip value of scale for dfq algo. Defaults to 10.
            eps (float, optional): small constant for numerical stability. Defaults to 1e-8.
        """
        first_weight_range, last_weight_range = self.prepare_weight_for_equalization(pair)

        if algo_type == 0:
            # dfq
            scale = torch.sqrt(last_weight_range / (first_weight_range + eps))
            scale = torch.clamp(scale, dfq_min_scale, dfq_max_scale)

        else:
            first_weight_range = torch.where(
                first_weight_range < first_weight_range.max() * self.channel_ratio,
                first_weight_range.max() * self.channel_ratio,
                first_weight_range,
            )
            last_weight_range = torch.where(
                last_weight_range < last_weight_range.max() * self.channel_ratio,
                last_weight_range.max() * self.channel_ratio,
                last_weight_range,
            )

            kernel_scale = first_weight_range.max() / (first_weight_range + eps)
            next_kernel_scale = last_weight_range.max() / (last_weight_range + eps)
            first_weight_act_range = op_act_channel_range[pair[0]]
            first_weight_act_range = torch.where(
                first_weight_act_range < 0.01,
                torch.tensor(0.01, device=first_weight_act_range.device, dtype=torch.float32),
                first_weight_act_range,
            )
            act_scale = first_weight_act_range.max() / (first_weight_act_range + eps)

            if algo_type == 1:
                scale = torch.min(kernel_scale, act_scale)
            elif algo_type == 2:
                kernel_scale = kernel_scale / next_kernel_scale
                act_scale = act_scale / next_kernel_scale
                scale = torch.min(kernel_scale, act_scale)
                scale = torch.min(scale, torch.tensor(ssd_min_scale, dtype=torch.float32, device=scale.device))
                scale /= scale.min()
                scale = torch.clamp(scale, 1.0, ssd_max_scale)
            else:
                kernel_scale = (kernel_scale / next_kernel_scale).sqrt()
                scale = (act_scale * kernel_scale).sqrt()
                scale = torch.clamp(scale, 1.0, ssd_max_scale)

        self.write_back(pair, scale)

    def build_observer_pair(self, pair: List[Operation]) -> Dict[Operation, OperationObserver]:
        observers = {}
        for operation in pair:
            if not isinstance(operation, QuantableOperation):
                continue
            observer = OperationObserver(operation=operation)
            observers[operation] = observer
        return observers

    def calibrate(
        self,
        pair: List[Operation],
        data_loader: Iterable,
        executor: BaseGraphExecutor,
        hooks: Dict[Operation, CalibrationHook],
        collate_fn: Callable,
        calib_steps: int,
    ):
        calib_step = 0
        for calib_epoch in range(ceil(calib_steps / len(data_loader))):
            for _, data in enumerate(data_loader):
                if collate_fn is not None:
                    data = collate_fn(data)
                # get the input of first op
                inputs = executor.forward(data, output_names=[pair[0].inputs[0].name])
                self.run_pair(pair, inputs, hooks)
                calib_step += 1
                if calib_step >= calib_steps:
                    break

    def calibration_passive_param(self, pair: List[Operation], scale_multiplier: float = 1.0):
        for op in pair:
            if not isinstance(op, QuantableOperation):
                continue
            if op.type in {"Conv", "ConvTranspose", "Gemm"}:
                if op.num_of_input == 3:
                    weight_config = op.config.input_quantization_config[1]
                    input_config = op.config.input_quantization_config[0]
                    weight_config = weight_config.dominated_by
                    input_config = input_config.dominated_by

                    bias_config = op.config.input_quantization_config[-1]
                    if bias_config.state != QuantizationStates.PASSIVE_INIT:
                        continue

                    bias_config.scale = weight_config.scale * input_config.scale * scale_multiplier
                    bias_config.state = QuantizationStates.PASSIVE
                    bias_config.offset = torch.zeros_like(bias_config.scale, dtype=torch.float)
                    assert not bias_config.policy.has_property(
                        QuantizationProperty.ASYMMETRICAL
                    ), "Negative parameter does not support ASYMMETRICAL quantization"

    def initiate_pair_state(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                for quant_config in op.config.input_quantization_config + op.config.output_quantization_config:
                    if quant_config.state == QuantizationStates.ACTIVATED:
                        quant_config.state = QuantizationStates.INITIAL
                    elif quant_config.state == QuantizationStates.PASSIVE:
                        quant_config.state = QuantizationStates.PASSIVE_INIT

    def dequantize_pair(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.dequantize(expire_device=None)

    def restore_quantize_state(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.restore_quantize_state(expire_device=None)

    def run_pair(
        self, pair: List[Operation], inputs: List[torch.Tensor], hooks: Dict[Operation, CalibrationHook] = {}
    ) -> List[torch.Tensor]:
        for op in pair:
            inputs = inputs + [param.value for param in op.parameters]
            if isinstance(op, QuantableOperation):
                input_configs = [_ for _ in op.config.input_quantization_config]
                assert len(inputs) == len(input_configs)
                inputs_quant = [self.quant_func(input, config) for input, config in zip(inputs, input_configs)]
                hook = hooks.get(op, None)
                if hook is not None:
                    hook.pre_forward_hook(inputs, inputs_quant, input_configs)
            else:
                inputs_quant = inputs

            f = OPERATION_FORWARD_TABLE[op.platform][op.type]
            outputs = f(op, inputs_quant)
            outputs = outputs if isinstance(outputs, (list, tuple)) else [outputs]

            if isinstance(op, QuantableOperation):
                output_configs = [_ for _ in op.config.output_quantization_config]
                outputs_quant = [self.quant_func(output, config) for output, config in zip(outputs, output_configs)]
                hook = hooks.get(op, None)
                if hook is not None:
                    hook.post_forward_hook(outputs, outputs_quant, output_configs)
                inputs = outputs_quant
            else:
                inputs = outputs
        return inputs

    # mse calculation for a list of tensors
    def calculate_mse(self, fp_res: List[torch.Tensor], quant_res: List[torch.Tensor]) -> torch.Tensor:
        losses = []
        for i in range(len(fp_res)):
            losses.append(torch_mean_square_error(fp_res[i], quant_res[i]))
        return torch.stack(losses).mean()

    @torch.no_grad()
    def test_ssd_loss(
        self,
        pair: List[Operation],
        executor: BaseGraphExecutor,
        data_loader: Iterable,
        collate_fn: Callable,
        calib_steps: int,
    ) -> float:
        observers = self.build_observer_pair(pair)
        hooks = {op: observers[op].hook for op in observers}
        self.calibrate(pair, data_loader, executor, hooks, collate_fn, calib_steps)
        for _, observer in observers.items():
            observer.render_quantization_config()
        pop_list = []
        for op, observer in observers.items():
            if all(
                [
                    type(var_observer) not in {TorchHistObserver}
                    for var_observer in observer._hook._observer_table.values()
                ]
            ):
                pop_list.append(op)
        for op in pop_list:
            observers.pop(op)
            hooks.pop(op)
        if len(hooks) > 0:
            self.calibrate(pair, data_loader, executor, hooks, collate_fn, calib_steps)
            for _, observer in observers.items():
                observer.render_quantization_config()
        self.calibration_passive_param(pair)
        # calculate loss
        loss = []
        for calib_epoch in range(ceil(calib_steps / len(data_loader))):
            for _, data in enumerate(data_loader):
                if collate_fn is not None:
                    data = collate_fn(data)
                # get the input of first op
                inputs = executor.forward(data, output_names=[pair[0].inputs[0].name])
                # dequant to get fp output
                self.dequantize_pair(pair)
                fp_output = self.run_pair(pair, inputs)
                # restore quant state to get quant output
                self.restore_quantize_state(pair)
                quant_output = self.run_pair(pair, inputs)
                # mse calculation
                loss.append(self.calculate_mse(fp_output, quant_output))
        return torch.stack(loss).mean().item()

    # maintain original parameter for restoration in case of a larger loss after equalization
    def collect_original_parameter(self, pair: List[Operation]) -> Dict[Variable, torch.Tensor]:
        original_weights = {}
        for op in pair:
            for var in op.inputs + op.outputs:
                if var.is_parameter:
                    original_weights[var] = var.value.clone()
        return original_weights

    # discard original fp value and store the current value as new fp value
    def store_parameter(self, pair: List[Operation]):
        for op in pair:
            if isinstance(op, QuantableOperation):
                op.store_parameter_value()

    # recover from maintained original parameters
    def recover_original_parameter(
        self,
        pair: List[Operation],
        original_weights: Dict[Variable, torch.Tensor],
    ):
        for op in pair:
            for var in op.inputs + op.outputs:
                if var.is_parameter:
                    var.value = original_weights[var]
        self.store_parameter(pair)

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        collate_fn: Callable,
        calib_steps: int,
        **kwargs,
    ) -> None:
        # restrain maximum img number used for loss checking
        batchsize = 1
        for data in dataloader:
            if collate_fn is not None:
                data = collate_fn(data)
            if isinstance(data, torch.Tensor):
                batchsize = data.shape[0]
            elif isinstance(data, (list, tuple)):
                for value in data:
                    if isinstance(value, torch.Tensor):
                        batchsize = value.shape[0]
                        break
            elif isinstance(data, dict):
                for key, value in data.items():
                    if isinstance(value, torch.Tensor):
                        batchsize = value.shape[0]
                        break
            break

        calib_steps = min(calib_steps, ceil(200 / batchsize))

        all_pairs = self.collect_all_pairs(graph)
        if self.layer_norm:
            self.layer_weight_norm(all_pairs)

        for i in range(self.iteration):
            logger.debug(f"DFQ/SSD Equalization Iteration {i + 1}/{self.iteration}")
            for _, pair in tqdm(
                enumerate(all_pairs),
                desc=f"SSD/DFQ Equalization Iteration {i+1}/{self.iteration}",
                total=len(all_pairs),
            ):
                logger.debug(f"Now Processing Pair {_ + 1}/{len(all_pairs)}: {'--'.join([op.name for op in pair])}")
                self.store_parameter(pair)

                logger.debug(f"Collecting Activation Range for Pair...")
                op_act_range = self.collect_activation_range(pair, executor, dataloader, collate_fn, calib_steps)
                logger.debug(f"Collecting Done!")

                original_weights = self.collect_original_parameter(pair)
                basic_loss = self.test_ssd_loss(pair, executor, dataloader, collate_fn, calib_steps)
                best_loss = basic_loss
                best_idx = -1

                # now apply equalization and estimate loss
                for algo in range(0, 4):
                    self.one_step_equalization(pair, op_act_range, algo)
                    self.store_parameter(pair)
                    self.initiate_pair_state(pair)
                    loss = self.test_ssd_loss(pair, executor, dataloader, collate_fn, calib_steps)
                    if algo == 0:
                        logger.debug(
                            f"DFQ Step, Loss Before Equalization {basic_loss} || Loss After Equalization {loss}"
                        )
                    else:
                        logger.debug(
                            f"SSD Algo {algo}, Loss Before Equalization {basic_loss} || Loss After Equalization {loss}"
                        )
                    if loss < basic_loss * self.loss_threshold and loss < best_loss:
                        best_idx = algo
                        best_loss = loss
                    self.recover_original_parameter(pair, original_weights)

                if best_idx >= 0:
                    if best_idx == 0:
                        logger.debug(
                            f"DFQ Step Activated, Loss Before Equalization {basic_loss} || Loss After Equalization {best_loss}"
                        )
                    else:
                        logger.debug(
                            f"SSD Algo {best_idx} Activated, Loss Before Equalization {basic_loss} || Loss After Equalization {best_loss}"
                        )
                    self.one_step_equalization(pair, op_act_range, best_idx)
                    self.store_parameter(pair)
                else:
                    logger.debug("SSD and DFQ Deactivated")
                self.initiate_pair_state(pair)



from typing import Any, List

from ....core import TargetPlatform
from ....IR import Operation
from ....IR.quantize import QuantableOperation

import torch


class TorchBackendContext:
    def __init__(self, executing_device: str) -> None:
        self.executing_device = executing_device


def ASSERT_NUM_OF_INPUT(
    op: Operation, values: List[torch.Tensor], min_num_of_input: int = -1, max_num_of_input: int = 99
):
    if min_num_of_input == max_num_of_input:
        if len(values) != min_num_of_input:
            raise ValueError(
                f"Can not feed value to operation {op.name}, "
                f"expects exact {min_num_of_input} inputs, however {len(values)} was given"
            )
    elif len(values) > max_num_of_input:
        raise ValueError(
            f"Too many input value for {op.name}, "
            f"expects {max_num_of_input} inputs at most, however {len(values)} was given"
        )
    elif len(values) < min_num_of_input:
        raise ValueError(
            f"Too few input value for {op.name}, "
            f"expects {min_num_of_input} inputs at least, however {len(values)} was given"
        )


def GET_ATTRIBUTE_FROM_OPERATION(op: Operation, attribute: str, compulsive: bool = False, default: Any = None):
    """Try to get an attribute from operation. If an attribute is compulsive,
    then operation must give a value of it, otherwise an error will be thrown.
    If an attribute is not compulsive, a default value will be given if
    operation.attributes do not holds a value of requesting attribute.

    Args:
        op (Operation): Operation instance.
        attribute (str): Attribute name.
        compulsive (bool): Whether is a compulsive attribute.
        default (Any, optional): [description]. default value of attribute.
    """
    if attribute in op.attributes:
        return op.attributes[attribute]
    else:
        if compulsive:
            raise KeyError(
                f"Operation {op.name} is supposed to have a value of attribute {attribute}. ",
                "However this value is missing from currecnt operation.",
            )
        else:
            return default


def GET_VALUE_FROM_INPUTS(values: list, idx: int) -> torch.Tensor:
    assert isinstance(idx, int)
    assert idx > 0
    if len(values) > idx:
        return values[idx]
    else:
        return None


def ASSERT_IS_QUANT_OP(op: QuantableOperation):
    if not isinstance(op, QuantableOperation):
        raise TypeError(f"Given Operation is expected as a QuantableOperation, however {type(op)} was given.")


def FORCE_CONVERT_DEVICE(value: torch.Tensor, device: str) -> torch.Tensor:
    # SET LOG HERE FOR DEBUG.
    # value = value.clone()
    return value.to(device=device, copy=True)


def VALUE_TO_EXECUTING_DEVICE(
    op: Operation, ctx: TorchBackendContext, values: List[torch.Tensor]
) -> List[torch.Tensor]:
    if ctx is None:
        device = values[0].device
    else:
        device = ctx.executing_device
    for idx, (plat, value) in enumerate(zip(op.socket.in_plat, values)):
        if value is None:
            continue
        if plat == TargetPlatform.SOI or op.platform == TargetPlatform.SOI:
            values[idx] = value.cpu()
        else:
            values[idx] = value.to(device)
    return values



from .default import (
    DEFAULT_BACKEND_TABLE,
    AveragePool_forward,
    AdaptiveAvgPool2d_forward,
    BatchNormalization_forward,
    Cast_forward,
    Clip_forward,
    Concat_forward,
    Constant_forward,
    ConstantOfShape_forward,
    Conv_forward,
    Eltwise_forward,
    Equal_forward,
    Expand_forward,
    Flatten_forward,
    Gather_forward,
    GatherND_forward,
    Gemm_forward,
    Greater_forward,
    MaxPool2d_forward,
    _NMS_forward,
    NonZero_forward,
    Range_forward,
    ReduceL2_forward,
    ReduceMax_forward,
    Reshape_forward,
    Resize_forward,
    ScatterElements_forward,
    ScatterND_forward,
    Shape_forward,
    Slice_forward,
    Softmax_forward,
    Squeeze_forward,
    Tile_forward,
    TopK_forward,
    Transpose_forward,
    UnaryEltwise_forward,
    Unsqueeze_forward,
    Where_forward,
)
from .base import TorchBackendContext



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
from .convert_to_fp16 import convert_to_fp16_onnx_model
from .dynamic_quantizer import dynamic_quantize_onnx_model
from .xslim import XSlimQuantizer



#!/usr/bin/env python3
# Copyright (c) 2025 SpacemiT. All rights reserved.
from typing import Sequence, Set, Tuple, Union

import onnx
import numpy as np
from onnxconverter_common import float16 as convert_float_to_float16
from xslim.logger import logger
import onnx_graphsurgeon as osg
from ..onnx_graph_helper import format_onnx_model
from ..onnxslim_pass import infer_onnx_model
from xslim.defs import XQUANT_CONFIG
from datetime import datetime


def legalize_fp16_graph(osg_graph: osg.Graph):
    for node in osg_graph.nodes:
        if node.op in {"Resize", "Upsample"}:
            for input_var in node.inputs:
                if isinstance(input_var, osg.Constant) and input_var.dtype == np.float16:
                    input_var.values = input_var.values.astype(np.float32)
        elif node.op in {"Cast"}:
            to_dtype = onnx.helper.tensor_dtype_to_np_dtype(node.attrs["to"])
            if node.outputs[0].dtype != to_dtype:
                node.attrs["to"] = onnx.helper.np_dtype_to_tensor_dtype(
                    node.outputs[0].dtype)
        elif node.op in {"Equal", "NotEqual", "Greater", "Less", "GreaterEqual", "LessEqual", "Add", "Sub", "Mul", "Div"}:
            if node.inputs[0].dtype == np.float16 and node.inputs[1].dtype != np.float16:
                if isinstance(node.inputs[1], osg.Constant):
                    node.inputs[1].values = node.inputs[1].values.astype(
                        np.float16)
                else:
                    raise RuntimeError(
                        "Unsupported op {}[{}] with fp16 inputs".format(node.op, node.name))
            elif node.inputs[0].dtype != np.float16 and node.inputs[1].dtype == np.float16:
                if isinstance(node.inputs[0], osg.Constant):
                    node.inputs[0].values = node.inputs[0].values.astype(
                        np.float16)
                else:
                    raise RuntimeError(
                        "Unsupported op {}[{}] with fp16 inputs".format(node.op, node.name))
        elif node.op in {"Range"}:
            remove_var = []
            add_var = []
            remove_idx = []
            for input_var in node.inputs:
                if isinstance(input_var, osg.Constant) and input_var.dtype == np.float16:
                    new_var = osg.Constant("{}_to_fp32".format(
                        input_var.name), input_var.values.astype(np.float32))
                    remove_idx.append(node.inputs.index(input_var))
                    remove_var.append(input_var)
                    add_var.append(new_var)

            for r_var, add_var, r_idx in zip(remove_var, add_var, remove_idx):
                node.inputs[r_idx] = add_var

    return osg_graph


def convert_to_fp16_onnx_model(
    file_or_model: Union[str, onnx.ModelProto],
    ignore_op_types_list: Sequence[str],
    ignore_node_names_list: Sequence[str],
    sim_en: bool = True,
):
    if isinstance(file_or_model, onnx.ModelProto):
        onnx_model = file_or_model
    elif isinstance(file_or_model, str):
        onnx_model = onnx.load(file_or_model)
    else:
        raise TypeError("type of file_or_model error, {} .vs str or modelproto".format(
            type(file_or_model)))

    model_opt = format_onnx_model(onnx_model, sim_en)

    logger.info("convert onnx model to fp16.")

    default_ignore_op_types = {"ArrayFeatureExtractor",
                               "Binarizer",
                               "CastMap",
                               "CategoryMapper",
                               "DictVectorizer",
                               "FeatureVectorizer",
                               "Imputer",
                               "LabelEncoder",
                               "LinearClassifier",
                               "LinearRegressor",
                               "Normalizer",
                               "OneHotEncoder",
                               "RandomUniformLike",
                               "RandomNormalLike",
                               "SVMClassifier",
                               "SVMRegressor",
                               "Scaler",
                               "TreeEnsembleClassifier",
                               "TreeEnsembleRegressor",
                               "ZipMap",
                               "NonMaxSuppression",
                               "TopK",
                               "RoiAlign",
                               "Range",
                               "CumSum"}

    try:
        model_fp16 = convert_float_to_float16.convert_float_to_float16(
            model_opt,
            keep_io_types=True,
            disable_shape_infer=False,
            op_block_list=list(
                default_ignore_op_types or set(ignore_op_types_list)),
            node_block_list=ignore_node_names_list,
        )
    except Exception as e:
        logger.info(f"FP16 Convert Failed!: {e}")
        raise

    model_fp16 = format_onnx_model(model_fp16)

    osg_graph = osg.import_onnx(model_fp16)
    osg_graph = legalize_fp16_graph(osg_graph)
    model_fp16 = osg.export_onnx(osg_graph)

    model_fp16 = format_onnx_model(model_fp16)

    model_fp16.producer_name = "xslim"
    export_time = model_fp16.metadata_props.add()
    export_time.key = "xslim_export_time"
    export_time.value = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
    xslim_version = model_fp16.metadata_props.add()
    xslim_version.key = "xslim_version"
    xslim_version.value = XQUANT_CONFIG.version

    return model_fp16



import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern


class LayernormPatternCase0Matcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input          input            0 2 reduce_mean_0 sub_0
            ReduceMean     reduce_mean_0    1 1 input sub_0
            Sub            sub_0            2 1 input reduce_mean_0 pow_0
            Pow            pow_0            2 1 sub_0 ? reduce_mean_1
            ReduceMean     reduce_mean_1    1 1 pow_0 add_0
            Add            add_0            2 1 reduce_mean_1 ? sqrt_0
            Sqrt           sqrt_0           1 1 add_0 div_0
            Div            div_0            2 1 sub_0 sqrt_0 mul_0
            Mul            mul_0            2 1 div_0 ? add_1
            Add            add_1            2 1 mul_0 ? output
            output         output           1 0 add_1
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionLayerNorm'."""
        return "FusionLayerNormCase0"

    def rewrite(self, opset=11):
        reduce_mean_0_node = self.reduce_mean_0
        reduce_mean_1_node = self.reduce_mean_1
        sub_0_node = self.sub_0
        pow_0_node = self.pow_0
        add_0_node = self.add_0
        div_node = self.div_0
        mul_0_node = self.mul_0
        add_1_node = self.add_1

        axes_0 = reduce_mean_0_node.attrs.get("axes", [])
        axes_1 = reduce_mean_1_node.attrs.get("axes", [])

        input_variable = reduce_mean_0_node.inputs[0]

        if (
            isinstance(pow_0_node.inputs[1], osg.Constant)
            and pow_0_node.inputs[1].values is not None
            and np.allclose(2.0, pow_0_node.inputs[1].values)
            and isinstance(add_0_node.inputs[1], osg.Constant)
            and isinstance(mul_0_node.inputs[1], osg.Constant)
            and isinstance(add_1_node.inputs[1], osg.Constant)
        ):

            norm_scale_constant = mul_0_node.inputs[1]
            norm_bias_constant = add_1_node.inputs[1]

            axis = max(axes_0)
            epsilon = float(add_0_node.inputs[1].values)

            if not len(norm_scale_constant.shape) == 1 or not len(norm_bias_constant.shape) == 1:
                return {}

            input_variable.outputs.remove(reduce_mean_0_node)
            input_variable.outputs.remove(sub_0_node)

            output_variable = add_1_node.outputs[0]
            output_variable.inputs.clear()

            return {
                reduce_mean_0_node.name: {
                    "op": "LayerNormalization",
                    "inputs": [input_variable, mul_0_node.inputs[1], add_1_node.inputs[1]],
                    "outputs": [output_variable],
                    "domain": None,
                    "attrs": {"axis": axis, "epsilon": epsilon},
                    "name": reduce_mean_0_node.name,
                }
            }
        else:
            return {}


class LayernormPatternCase1Matcher(PatternMatcher):
    def __init__(self, priority):
        pattern = Pattern(
            """
            input          input            0 2 reduce_mean_0 sub_0
            ReduceMean     reduce_mean_0    1 1 input sub_0
            Sub            sub_0            2 1 input reduce_mean_0 pow_0
            Pow            pow_0            2 1 sub_0 ? reduce_mean_1
            ReduceMean     reduce_mean_1    1 1 pow_0 add_0
            Add            add_0            2 1 reduce_mean_1 ? sqrt_0
            Sqrt           sqrt_0           1 1 add_0 div_0
            Div            div_0            2 1 sub_0 sqrt_0 mul_0
            Mul            mul_0            2 1 ? div_0 add_1
            Add            add_1            2 1 mul_0 ? output
            output         output           1 0 add_1
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionLayerNorm'."""
        return "FusionLayerNormCase1"

    def rewrite(self, opset=11):
        reduce_mean_0_node = self.reduce_mean_0
        reduce_mean_1_node = self.reduce_mean_1
        sub_0_node = self.sub_0
        pow_0_node = self.pow_0
        add_0_node = self.add_0
        div_node = self.div_0
        mul_0_node = self.mul_0
        add_1_node = self.add_1

        axes_0 = reduce_mean_0_node.attrs.get("axes", [])
        axes_1 = reduce_mean_1_node.attrs.get("axes", [])

        input_variable = reduce_mean_0_node.inputs[0]

        if (
            isinstance(pow_0_node.inputs[1], osg.Constant)
            and pow_0_node.inputs[1].values is not None
            and np.allclose(2.0, pow_0_node.inputs[1].values)
            and isinstance(add_0_node.inputs[1], osg.Constant)
            and isinstance(mul_0_node.inputs[0], osg.Constant)
            and isinstance(add_1_node.inputs[1], osg.Constant)
        ):

            norm_scale_constant = mul_0_node.inputs[0]
            norm_bias_constant = add_1_node.inputs[1]

            axis = max(axes_0)
            epsilon = float(add_0_node.inputs[1].values)

            append_ops = {}

            input_variable.outputs.remove(reduce_mean_0_node)
            input_variable.outputs.remove(sub_0_node)

            output_variable = add_1_node.outputs[0]
            output_variable.inputs.clear()

            ln_input = input_variable
            ln_output = output_variable
            if not len(norm_scale_constant.shape) == 1 or not len(norm_bias_constant.shape) == 1:
                ndim = len(norm_scale_constant.shape) + axis
                permute_in = [i for i in range(axis)] + [i for i in range(axis + 1, ndim)] + [axis]
                permute_out = [i for i in range(axis)] + [ndim - 1] + [i - 1 for i in range(axis + 1, ndim)]

                ln_input = osg.Variable("{}_ln_input".format(reduce_mean_0_node.name), np.float32)
                ln_output = osg.Variable("{}_ln_output".format(reduce_mean_0_node.name), np.float32)

                append_ops["{}_permute_in".format(reduce_mean_0_node.name)] = {
                    "op": "Transpose",
                    "inputs": [input_variable],
                    "outputs": [ln_input],
                    "domain": None,
                    "attrs": {"perm": permute_in},
                    "name": "{}_permute_in".format(reduce_mean_0_node.name),
                }

                append_ops["{}_permute_out".format(reduce_mean_0_node.name)] = {
                    "op": "Transpose",
                    "inputs": [ln_output],
                    "outputs": [output_variable],
                    "domain": None,
                    "attrs": {"perm": permute_out},
                    "name": "{}_permute_out".format(reduce_mean_0_node.name),
                }

                axis = -1
                mul_0_node.inputs[0].values = norm_scale_constant.values.reshape(-1)
                add_1_node.inputs[1].values = norm_bias_constant.values.reshape(-1)

            append_ops[reduce_mean_0_node.name] = {
                "op": "LayerNormalization",
                "inputs": [ln_input, mul_0_node.inputs[0], add_1_node.inputs[1]],
                "outputs": [ln_output],
                "domain": None,
                "attrs": {"axis": axis, "epsilon": epsilon},
                "name": reduce_mean_0_node.name,
            }

            return append_ops
        else:
            return {}


register_fusion_pattern(LayernormPatternCase0Matcher(1))
register_fusion_pattern(LayernormPatternCase1Matcher(1))



#!/usr/bin/env python3
# Copyright (c) 2023 SpacemiT. All rights reserved.
import functools
from collections import deque
from typing import Callable, Dict, Iterable, List, Sequence, Set, Tuple, Union

import torch
from tqdm import tqdm

from ..defs import BIAS_CORRECTION_INTERST_TYPE, COMPUTING_OP, OBSERVER_MAX_BIAS_VAL, PASSIVE_OPERATIONS, XQUANT_CONFIG
from ..ppq_decorator import (
    BaseGraph,
    Operation,
    OperationObserver,
    QuantableOperation,
    QuantizationOptimizationPass,
    QuantizationStates,
    TorchExecutor,
    Variable,
    ppq_common,
    torch_cosine_similarity,
    torch_mean_square_error,
    torch_snr_error,
)
from .observer import TorchXSlimObserver
from .refine import PassiveParameterBakingPass
from .training import LearnedStepSizePassDecorator, XSlimBlockBuilder, XSlimTrainableBlock


class BlockWiseCalibrationDataset(torch.utils.data.Dataset):
    def __init__(self, dataloader_cache: Sequence[dict]):
        # torch.save(dataloader_cache, "calib_dataloader_cache")
        self.__data_cache = dataloader_cache
        # self.__data_cache = torch.load("calib_dataloader_cache", mmap=True, map_location=map_location)

    def __len__(self):
        return len(self.__data_cache)

    def __getitem__(self, idx):
        return self.__data_cache[idx]

    def update_data(self, idx, name, data):
        if data.device.type == "cuda":
            mem_free, mem_all = torch.cuda.mem_get_info()
            mem_free_ratio = mem_free / mem_all
            if mem_free_ratio < 0.2:
                data = data.to("cpu")
        self.__data_cache[idx].update({name: data})


class RuntimeBlockWiseCalibrationPass(QuantizationOptimizationPass):
    def __init__(
        self,
        method: str = None,
        override: bool = False,
        calib_steps: int = 32,
        block_wise: bool = True,
        fintune_epoch: int = 2,
        auto_finetune_level: int = 1,
    ) -> None:
        self.name = "XSlim Runtime Calibration Pass(BlockWise)"
        self._method = method
        self._collate_fn = None
        self._calib_steps = calib_steps
        self._override = override
        self._block_wise = block_wise
        self._fintune_epoch = fintune_epoch
        self._block_wise_loss = []
        self._auto_finetune_level = auto_finetune_level

    def split_graph_into_blocks(
        self,
        graph: BaseGraph,
        executing_order: List[Operation],
    ) -> List[XSlimTrainableBlock]:
        block_builder = XSlimBlockBuilder(graph=graph, topo_order=executing_order)
        return block_builder.build()

    def forward_trainable_block(
        self,
        block: XSlimTrainableBlock,
        dataset: BlockWiseCalibrationDataset,
        executor: TorchExecutor,
        executor_hook: dict,
        calib_steps: int,
        extern_output_var_hooks: Dict[str, Callable] = None,
        collect_dataloader_cache: bool = False,
    ):
        operation_cache = block.rps
        if extern_output_var_hooks is None:
            extern_output_var_hooks = {}

        block_input_names = list(block.in_var_names)

        single_data = dataset[0]
        for var_name in block_input_names:
            if var_name not in single_data:
                raise RuntimeError("missing input variable {} for block".format(var_name))

        for idx, batch_data in tqdm(enumerate(dataset), desc="Runtime Calibration Single Block Collect", disable=True):
            inputs_feed = {
                var_name: batch_data[var_name].to(executor._executing_context.executing_device)
                for var_name in block_input_names
            }
            output_names = list(block.out_var_names | set(extern_output_var_hooks.keys()))
            outputs = executor._TorchExecutor__forward(
                inputs_feed,
                operation_cache,
                output_names=output_names,
                hooks=executor_hook,
            )
            for o_var, o_name in zip(outputs, output_names):
                if o_name in block.out_var_names and collect_dataloader_cache:
                    dataset.update_data(idx, o_name, o_var)
                    batch_data[o_name] = o_var
                if o_name in extern_output_var_hooks:
                    extern_output_var_hooks[o_name](o_name, o_var, batch_data)

    def clean_dataloader_cache(
        self,
        block: XSlimTrainableBlock,
        dataset: BlockWiseCalibrationDataset,
        var_to_operations: dict,
    ):
        remove_op_set = set(block.rps)
        remove_ovarnames = []
        for o_name, o_var in dataset[0].items():
            if o_name in var_to_operations:
                var_to_operations[o_name] -= remove_op_set
                if len(var_to_operations[o_name]) == 0:
                    remove_ovarnames.append(o_name)
        for data in dataset:
            for o_name in remove_ovarnames:
                data.pop(o_name)

    def calib_single_block(
        self,
        block: XSlimTrainableBlock,
        dataset: BlockWiseCalibrationDataset,
        var_to_operations: dict,
        executor: TorchExecutor,
        calib_steps: int,
    ):
        bias_op_var_names = dict()
        bias_fp_cache = dict()
        operation_observer_cache = []
        extern_output_var_hooks = {}

        def __collect_bias(o_name: str, o_var: torch.Tensor, batch_data: Dict[str, torch.Tensor]):
            reduce_bias = self.collect_bias(o_var, bias_op_var_names[o_name])
            if o_name not in bias_fp_cache:
                bias_fp_cache[o_name] = []
            bias_fp_cache[o_name].append(reduce_bias)

        for operation in block.rps:
            if isinstance(operation, QuantableOperation):
                operation_observer = OperationObserver(operation=operation, monitor_parameter=False)
                operation_observer_cache.append(operation_observer)
                if operation.type in BIAS_CORRECTION_INTERST_TYPE and operation.num_of_parameter == 2:
                    bias_op_var_names[operation.outputs[0].name] = operation
                    extern_output_var_hooks[operation.outputs[0].name] = __collect_bias

        ob_table_num = sum([len(ob.hook._observer_table) for ob in operation_observer_cache])
        hooks = {ob._operation.name: ob.hook for ob in operation_observer_cache}
        if ob_table_num == 0:
            hooks = {}

        not_has_hist_ob = True
        for ob in operation_observer_cache:
            if not_has_hist_ob:
                not_has_hist_ob = all(
                    [
                        not isinstance(var_observer, TorchXSlimObserver)
                        for var_observer in ob._hook._observer_table.values()
                    ]
                )

        with torch.no_grad():
            self.forward_trainable_block(block, dataset, executor, hooks, calib_steps, extern_output_var_hooks, True)

        if len(hooks) == 0:
            return

        for ob in operation_observer_cache:
            ob.render_quantization_config()
            ob.report()

        if not_has_hist_ob:
            pass
        else:
            with torch.no_grad():
                self.forward_trainable_block(block, dataset, executor, hooks, calib_steps)
            for ob in operation_observer_cache:
                ob.render_quantization_config()
                ob.report()

        if self._auto_finetune_level >= 1:
            self.block_bias_correct(
                block,
                dataset,
                executor,
                calib_steps,
                bias_op_var_names,
                bias_fp_cache,
            )

        loss_dict = self.compute_block_loss(block, dataset, executor, calib_steps)
        self._block_wise_loss.append(
            self.create_block_loss_info(
                block, loss_dict["mse"], loss_dict["snr"], loss_dict["mean"], loss_dict["cosine"]
            )
        )
        self._block_wise_loss[-1]["block_idx"] = len(self._block_wise_loss)
        self.clean_dataloader_cache(block, dataset, var_to_operations)

    def format_tqc(self, operaion: Operation):
        if not isinstance(operaion, QuantableOperation):
            return

        for tqc, var in operaion.config_with_variable:
            if tqc.dominated_by is tqc and tqc.scale is not None:
                if torch.any(tqc.scale < 0):
                    raise RuntimeError("tqc.scale for {} < 0, you can change fintune_level <= 1.".format(operaion.name))

        PassiveParameterBakingPass.passive_parameters_quant(operaion)
        PassiveParameterBakingPass.passive_bias_quant(operaion)

    def compute_block_loss(
        self,
        block: XSlimTrainableBlock,
        dataset: BlockWiseCalibrationDataset,
        executor: TorchExecutor,
        calib_steps: int,
    ) -> dict:
        block_output_names = list(block.out_var_names)
        mse_losses = {o_name: 0 for o_name in block_output_names}
        snr_losses = {o_name: 0 for o_name in block_output_names}
        cosine = {o_name: 0 for o_name in block_output_names}
        means = {o_name: 0 for o_name in block_output_names}

        def __collect_loss(o_name: str, o_var: torch.Tensor, batch_data: Dict[str, torch.Tensor]):
            exe_device = executor._executing_context.executing_device
            ref_var = batch_data[o_name]
            ref_var = ref_var.to(exe_device)
            o_var = o_var.to(exe_device)
            if ref_var.dtype in {torch.float32, torch.float64, torch.float16}:
                batch_mse_loss = torch_mean_square_error(o_var, ref_var)
                batch_snr_loss = torch_snr_error(o_var, ref_var)
                batch_cosine = torch_cosine_similarity(o_var, ref_var)
                means[o_name] += ref_var.abs().mean().detach().item()
            else:
                batch_mse_loss = torch.tensor([0], dtype=torch.float32, device=exe_device)
                batch_snr_loss = torch.tensor([0], dtype=torch.float32, device=exe_device)
                batch_cosine = torch.tensor([1], dtype=torch.float32, device=exe_device)
                means[o_name] += torch.tensor([0], dtype=torch.float32, device=exe_device)
            mse_losses[o_name] += batch_mse_loss.detach().item()
            snr_losses[o_name] += batch_snr_loss.detach().item()
            cosine[o_name] += batch_cosine.detach().item()

        extern_output_var_hooks = {o_name: __collect_loss for o_name in block_output_names}

        with torch.no_grad():
            self.forward_trainable_block(block, dataset, executor, {}, calib_steps, extern_output_var_hooks)

        for o_name in mse_losses:
            mse_losses[o_name] /= calib_steps
            snr_losses[o_name] /= calib_steps
            means[o_name] /= calib_steps
            cosine[o_name] /= calib_steps

        return {"mse": mse_losses, "snr": snr_losses, "cosine": cosine, "mean": means}

    def collect_bias(self, output: torch.Tensor, op: Operation) -> torch.Tensor:
        if output.ndim < 1:
            raise ValueError("Forward value has an unexpected dimension.")
        op_type = op.type
        if op_type in {"Conv", "ConvTranspose", "InstanceNormalization", "GroupNormalization"}:
            # for convolution layer, bias always been added on axis 1
            reduce_dims = [i for i in range(output.ndim) if i != 1]
            return torch.mean(output, dim=reduce_dims).unsqueeze(0)
        elif op_type in {"Gemm"}:
            reduce_dims = [i for i in range(output.ndim) if i != (output.ndim - 1)]
            return torch.mean(output, dim=(0,)).unsqueeze(0)
        elif op_type in {"LayerNormalization"}:
            axis = op.attributes.get("axis", -1)
            if axis < 0:
                axis = output.ndim + axis
            reduce_dims = [i for i in range(axis)]
            return torch.mean(output, dim=reduce_dims).unsqueeze(0)
        else:
            raise TypeError(f"Unsupported Operation type: {op_type}")

    def block_bias_correct(
        self,
        block: XSlimTrainableBlock,
        dataset: BlockWiseCalibrationDataset,
        executor: TorchExecutor,
        calib_steps: int,
        bias_op_var_names: dict,
        bias_fp_cache: dict,
    ) -> None:
        if len(bias_op_var_names) == 0:
            return
        operation_cache = [operation for operation in block.rps]
        bias_quant_cache = {}
        output_names = list(bias_op_var_names.keys())
        for batch_data in tqdm(dataset, desc="Runtime Calibration Single Block Finetune", disable=True):
            inputs_feed = {
                var_name: batch_data[var_name].to(executor._executing_context.executing_device)
                for var_name in block.in_var_names
            }
            outputs = executor.partial_graph_forward(
                operations=operation_cache, feed_dict=inputs_feed, output_names=list(bias_op_var_names.keys())
            )
            for o_var, o_name in zip(outputs, output_names):
                reduce_bias = self.collect_bias(o_var, bias_op_var_names[o_name])
                if o_name not in bias_quant_cache:
                    bias_quant_cache[o_name] = [reduce_bias]
                else:
                    bias_quant_cache[o_name].append(reduce_bias)

        def get_bias_valid_mean(value: torch.Tensor):
            target_device = value.device
            percentile = 0.999
            value = value.cpu()
            batch, channel = value.size()
            value = value.permute(1, 0)
            new_bias = []

            min_idx, max_idx = int(batch * (1 - percentile)), int(batch * (percentile))
            min_idx = max(0, min_idx) + 1
            max_idx = min(max_idx, batch - 1) + 1

            for i in range(channel):
                temp_val = value[i].flatten()
                _min = torch.kthvalue(temp_val, k=min_idx, dim=0)[0].item()
                _max = torch.kthvalue(temp_val, k=max_idx, dim=0)[0].item()
                temp_val = temp_val[torch.where(temp_val >= _min)]
                temp_val = temp_val[torch.where(temp_val <= _max)]
                new_bias.append(temp_val.mean().item())
            return torch.tensor(new_bias, dtype=value.dtype, device=target_device)

        for o_name in bias_op_var_names:
            DC_term_fp = bias_fp_cache[o_name]
            DC_term_qt = bias_quant_cache[o_name]

            if len(DC_term_fp) == 0 or len(DC_term_qt) == 0:
                continue
            DC_term_fp = get_bias_valid_mean(torch.cat(DC_term_fp, axis=0))
            DC_term_qt = get_bias_valid_mean(torch.cat(DC_term_qt, axis=0))
            bias_error = DC_term_fp - DC_term_qt
            bias_op_var_names[o_name].inputs[-1].value += bias_error

    def create_block_loss_info(
        self, block: XSlimTrainableBlock, mse_losses: dict, snr_losses: dict, means: dict, cosines: dict
    ):
        return {
            "start_op": "[{}]{}".format(block.sp.type, block.sp.name),
            "end_op": "[{}]{}".format(block.ep.type, block.ep.name),
            "snr": max([v for k, v in snr_losses.items()]),
            "mse": max([v for k, v in mse_losses.items()]),
            "cosine": min([v for k, v in cosines.items()]),
            "block": block,
        }

    def finetune(
        self,
        graph: BaseGraph,
        blocks: Sequence[XSlimTrainableBlock],
        dataloader: Iterable,
        executor: TorchExecutor,
        calib_steps: int = 32,
        collate_fn: Callable = None,
    ):
        lsq_optimizer = LearnedStepSizePassDecorator(is_scale_trainable=True, lr=0.0001)
        for block in tqdm(blocks, desc="Runtime Calibration Single Block Finetune"):
            qt_inputs, fp_outputs = lsq_optimizer.collect(
                graph=graph,
                block=block,
                executor=executor,
                dataloader=dataloader,
                collate_fn=collate_fn,
                collecting_device=executor._executing_context.executing_device,
            )

            pre_loss, post_loss = lsq_optimizer.finetune(
                steps=calib_steps * self._fintune_epoch,
                learning_rate=lsq_optimizer.lr,
                block=block,
                qt_inputs=qt_inputs,
                fp_outputs=fp_outputs,
                executor=executor,
            )

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: TorchExecutor,
        calib_steps: int = 32,
        collate_fn: Callable = None,
        **kwargs,
    ) -> None:
        self._collate_fn = collate_fn
        self._calib_steps = calib_steps
        self._block_wise_loss.clear()
        assert calib_steps >= 10, (
            "Insufficient Calibration Detected, to better quantize your network, "
            "more calibration steps is demonded, we strongly recommend you to prepare more calibration data "
            "and more calibration steps is perferred here. (at least 10)"
        )
        assert calib_steps <= 1000, (
            "Calibration steps is too large, xslim is capable for quantizing your network within 10-1000 "
            "calibration steps. More calibraiton steps will greatly delay ppq's calibration procedure. "
            "Reset your calib_steps parameter please."
        )
        if self._override:
            for operation in graph.operations.values():
                if not isinstance(operation, QuantableOperation):
                    continue

                for config, var in operation.config_with_variable:
                    if (
                        not var.is_parameter
                        and config.state == QuantizationStates.ACTIVATED
                        and config.dominated_by == config
                    ):
                        config.state = QuantizationStates.INITIAL

        single_graph_input_name = None
        dataloader_cache = []
        for k, v in graph.inputs.items():
            single_graph_input_name = k

        calib_step = 0
        for data in dataloader:
            data = self._collate_fn(data)
            if isinstance(data, torch.Tensor) and len(graph.inputs) == 1:
                if self._collate_fn is not None:
                    dataloader_cache.append({single_graph_input_name: data})
            elif isinstance(data, dict):
                dataloader_cache.append(data)
            else:
                raise TypeError(type(data))

            calib_step += 1
            if calib_step >= self._calib_steps:
                break

        var_to_operations = dict()
        for var_name, var in graph.variables.items():
            if not var.is_parameter:
                var_to_operations[var.name] = set([_ for _ in var.dest_ops])

        topo_sort_ops = graph.topological_sort()

        blockwise_dataset = BlockWiseCalibrationDataset(dataloader_cache)

        if self._block_wise:
            op_blocks = self.split_graph_into_blocks(graph, topo_sort_ops)
            for block in tqdm(op_blocks, desc="Runtime Calibration(BlockWise)"):
                self.calib_single_block(block, blockwise_dataset, var_to_operations, executor, calib_step)

            self._block_wise_loss = sorted(self._block_wise_loss, key=lambda x: x["mse"], reverse=True)
            if self._auto_finetune_level >= 2:
                _auto_finetune_blocks = 10 if self._auto_finetune_level >= 3 else 5
                finetune_blocks_info = [loss_info for loss_info in self._block_wise_loss[:_auto_finetune_blocks]]
                finetune_blocks_info = sorted(finetune_blocks_info, key=lambda x: x["block_idx"])
                self.finetune(
                    graph,
                    [loss_info["block"] for loss_info in finetune_blocks_info],
                    dataloader,
                    executor,
                    calib_step,
                    collate_fn,
                )
        else:
            raise NotImplementedError("block wise only.")

        for op in topo_sort_ops:
            self.format_tqc(op)



import numpy as np
import onnxslim.third_party.onnx_graphsurgeon as osg
from onnxslim.core.pattern import Pattern, PatternMatcher
from onnxslim.core.pattern.registry import register_fusion_pattern


class SwishPatternMatcher(PatternMatcher):
    def __init__(self, priority):
        """Initializes a `SwishPatternMatcher` to identify and fuse Swish patterns in a computational graph."""
        pattern = Pattern(
            """
            input       input  0 2 sig_0 mul_0
            Sigmoid     sig_0  1 1 input mul_0
            Mul         mul_0  2 1 input sig_0 output
            output      output 1 0 mul_0
            """
        )
        super().__init__(pattern, priority)

    @property
    def name(self):
        """Returns the name of the fusion pattern, 'FusionSwish'."""
        return "FusionSwish"

    def rewrite(self, opset=24):
        """Rewrite the computation graph pattern to fuse Swish operations."""
        input_variable = self.sig_0.inputs[0]
        output_variable = self.mul_0.outputs[0]

        input_variable.outputs.remove(self.sig_0)
        input_variable.outputs.remove(self.mul_0)
        output_variable.inputs.clear()

        alpha = 1.0

        return {
            self.mul_0.name: {
                "op": "Swish",
                "inputs": [input_variable],
                "outputs": [output_variable],
                "domain": None,
                "name": self.mul_0.name,
                "attrs": {"alpha": alpha},
            }
        }


register_fusion_pattern(SwishPatternMatcher(1))
